# -*- coding: utf-8 -*-
import re
import sys

from underthesea.pipeline.text_normalize import token_normalize
from underthesea.pipeline.text_normalize.character_normalize import normalize_characters_in_text

UPPER = "[" + "".join([
    "A-Z",
    "ÀÁẢÃẠ",
    "ĂẰẮẲẴẶ",
    "ÂẦẤẨẪẬ",
    "Đ",
    "ÈÉẺẼẸ",
    "ÊỀẾỂỄỆ",
    "ÌÍỈĨỊ",
    "ÒÓỎÕỌ",
    "ÔỒỐỔỖỘ",
    "ƠỜỚỞỠỢ",
    "ÙÚỦŨỤ",
    "ƯỪỨỬỮỰ",
    "ỲÝỶỸỴ"
]) + "]"
LOWER = UPPER.lower()
W = "[" + UPPER[1:-1] + LOWER[1:-1] + "]"  # upper and lower

VIETNAMESE_CHARACTERS_UPPER = "ABCDEFGHIJKLMNOPQRSTUVXYZ" \
                              + "ÀÁẢÃẠ" + "ĂẰẮẲẴẶ" + "ÂẦẤẨẪẬ" \
                              + "Đ" \
                              + "ÈÉẺẼẸ" + "ÊỀẾỂỄỆ" \
                              + "ÌÍỈĨỊ" \
                              + "ÒÓỎÕỌ" + "ÔỒỐỔỖỘ" + "ƠỜỚỞỠỢ" \
                              + "ÙÚỦŨỤ" + "ƯỪỨỬỮỰ" \
                              + "ỲÝỶỸỴ"
VIETNAMESE_CHARACTERS_LOWER = VIETNAMESE_CHARACTERS_UPPER.lower()
VIETNAMESE_VOWELS_UPPER = "AEIOU" \
                          + "ÀÁẢÃẠ" + "ĂẰẮẲẴẶ" + "ÂẦẤẨẪẬ" \
                          + "ÈÉẺẼẸ" + "ÊỀẾỂỄỆ" \
                          + "ÌÍỈĨỊ" \
                          + "ÒÓỎÕỌ" + "ÔỒỐỔỖỘ" + "ƠỜỚỞỠỢ" \
                          + "ÙÚỦŨỤ" + "ƯỪỨỬỮỰ" \
                          + "ỲÝỶỸỴ"
VIETNAMESE_VOWELS_LOWER = VIETNAMESE_VOWELS_UPPER.lower()

#################################################
# PRIORITY 1                                    #
#################################################
specials = [
    r"=\>",
    r"==>",
    r"->",
    r"\.{2,}",
    r"-{2,}",
    r">>",
    r"\d+x\d+",  # dimension: 3x4
    r"v\.v\.\.\.",
    r"v\.v\.",
    r"v\.v",
    r"°[CF]"
]
specials = "(?P<special>(" + "|".join(specials) + "))"

abbreviations = [
    r"[A-ZĐ]+&[A-ZĐ]+",  # & at middle of word (e.g. H&M)
    r"T\.Ư",  # dot at middle of word
    f"{UPPER}+(?:\.{W}+)+\.?",
    f"{W}+['’]{W}+",  # ' ’ at middle of word
    # e.g. H'Mông, xã N’Thôn Hạ
    r"[A-ZĐ]+\.(?!$)",  # dot at the end of word
    r"Tp\.",
    r"Mr\.", "Mrs\.", "Ms\.",
    r"Dr\.", "ThS\.", "Th.S", "Th.s",
    r"e-mail",  # - at middle of word
    r"\d+[A-Z]+\d*-\d+",  # vehicle plates
    # e.g. 43H-0530
    r"NĐ-CP"
]
abbreviations = "(?P<abbr>(" + "|".join(abbreviations) + "))"

#################################################
# PRIORITY 2                                    #
#################################################

# urls pattern from nltk
# https://www.nltk.org/_modules/nltk/tokenize/casual.html
# with Vu Anh's modified to match fpt protocol
url = r"""             # Capture 1: entire matched URL
  (?:
  (ftp|http)s?:               # URL protocol and colon
    (?:
      /{1,3}            # 1-3 slashes
      |                 #   or
      [a-z0-9%]         # Single letter or digit or '%'
                        # (Trying not to match e.g. "URI::Escape")
    )
    |                   #   or
                        # looks like domain name followed by a slash:
    [a-z0-9.\-]+[.]
    (?:[a-z]{2,13})
    /
  )
  (?:                                  # One or more:
    [^\s()<>{}\[\]]+                   # Run of non-space, non-()<>{}[]
    |                                  #   or
    \([^\s()]*?\([^\s()]+\)[^\s()]*?\) # balanced parens, one level deep: (...(...)...)
    |
    \([^\s]+?\)                        # balanced parens, non-recursive: (...)
  )+
  (?:                                  # End with:
    \([^\s()]*?\([^\s()]+\)[^\s()]*?\) # balanced parens, one level deep: (...(...)...)
    |
    \([^\s]+?\)                        # balanced parens, non-recursive: (...)
    |                                  #   or
    [^\s`!()\[\]{};:'".,<>?«»“”‘’]     # not a space or one of these punct chars
  )
  |                        # OR, the following to match naked domains:
  (?:
    (?<!@)                 # not preceded by a @, avoid matching foo@_gmail.com_
    [a-z0-9]+
    (?:[.\-][a-z0-9]+)*
    [.]
    (?:[a-z]{2,13})
    \b
    /?
    (?!@)                  # not succeeded by a @,
                           # avoid matching "foo.na" in "foo.na@example.com"
  )
"""
url = "(?P<url>" + url + ")"

email = r"[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+"
email = "(?P<email>" + email + ")"

phone = [
    r"\d{2,}-\d{3,}-\d{3,}"  # e.g. 03-5730-2357
    # very careful, it's easy to conflict with datetime
]
phone = "(?P<phone>(" + "|".join(phone) + "))"

datetime = [
    # date
    r"\d{1,2}\/\d{1,2}\/\d+",  # e.g. 02/05/2014
    r"\d{1,2}\/\d{1,4}",  # e.g. 02/2014
    #   [WIP] conflict with number 1/2 (a half)
    r"\d{1,2}-\d{1,2}-\d+",  # e.g. 02-03-2014
    r"\d{1,2}-\d{1,4}",  # e.g. 08-2014
    #   [WIP] conflict with range 5-10 (from 5 to 10)
    r"\d{1,2}\.\d{1,2}\.\d+",  # e.g. 20.08.2014
    r"\d{4}\/\d{1,2}\/\d{1,2}",  # e.g. 2014/08/20
    r"\d{2}:\d{2}:\d{2}"  # time
    # e.g. 10:20:50 (10 hours, 20 minutes, 50 seconds)
]
datetime = "(?P<datetime>(" + "|".join(datetime) + "))"

name = [
    r"\d+[A-Z]+\d+",
    r"\d+[A-Z]+"  # e.g. 4K
]
name = "(?P<name>(" + "|".join(name) + "))"

number = [
    r"\d+(?:\.\d+)+,\d+",  # e.g. 4.123,2
    r"\d+(?:\.\d+)+",  # e.g. 60.542.000
    r"\d+(?:,\d+)+",  # e.g. 100,000,000
    r"\d+(?:[\.,_]\d+)?",  # 123
]
number = "(?P<number>(" + "|".join(number) + "))"

emoji = [
    r":\)\)*",
    r"=\)\)+",
    r"♥‿♥",
    r":D+(?=\s)",  # :D
    r":D+(?=$)",  # special e.g. Đạo diễn :Dương Tuấn Anh
    r"<3"  # heart
]
emoji = "(?P<emoji>(" + "|".join(emoji) + "))"

punct = [
    r"\.",
    r"\,",
    r"\(",
    r"\)",
    r"ʺ"  # Modifier Letter Double Prime symbol (U+02BA)
]
punct = "(?P<punct>(" + "|".join(punct) + "))"

#################################################
# PRIORITY 3                                    #
#################################################
word = r"(?P<word>\w+)"

word_hyphen = [
    r"(?<=\b)\w+\-[\w+-]*\w+"  # before word_hyphen must be word boundary
    # case to notice: 1.600m-2.000m
]
word_hyphen = "(?P<word_hyphen>(" + "|".join(word_hyphen) + "))"

symbol = [
    r"\+",
    r"×",
    r"-",
    r"÷",
    r":+",
    r"%",
    r"%",
    r"\$",
    r"\>",
    r"\<",
    r"=",
    r"\^",
    r"_",
    r":+"
]
symbol = "(?P<sym>(" + "|".join(symbol) + "))"

non_word = r"(?P<non_word>[^\w\s])"

# Caution: order is very important for regex
regex_patterns = [
    specials,  # Priority 1
    abbreviations,
    url,  # Priority 2
    email,
    phone,
    datetime,  # datetime must be before number
    name,
    number,
    emoji,
    punct,
    word_hyphen,  # Priority 3
    word,
    symbol,
    non_word  # non_word must be last
]
recompile_regex_patterns = False
regex_patterns_combine = r"(" + "|".join(regex_patterns) + ")"
if sys.version_info < (3, 0):
    regex_patterns_combine = regex_patterns_combine.decode('utf-8')
patterns = re.compile(regex_patterns_combine, re.VERBOSE | re.UNICODE)

# regex pattern to match extacted word "Viện nghiên cứu" or "chien luoc"
# and not match to other multi tokens word, such as "quoc gia"
pattern = re.compile(r"(\w+)(?:\s+)(\w+)", re.UNICODE)


def extract_match(m):
    for k, v in m.groupdict().items():
        if v is not None:
            return v, k


def tokenize(text, format=None, tag=False, use_character_normalize=True, use_token_normalize=True, fixed_words=[]):
    """
    tokenize text for word segmentation

    Args:
        use_token_normalize: use token normalize or not
        use_character_normalize: use character normalize or not
        tag: return token with tag or not
        format: format of result, default is None
    """
    global recompile_regex_patterns
    global patterns
    if len(fixed_words) > 0:
        compiled_fixed_words = [re.sub(" ", "\ ", fixed_word) for fixed_word in fixed_words]
        fixed_words_pattern = "(?P<fixed_words>\\b" + "\\b|\\b".join(compiled_fixed_words) + "\\b)"
        merged_regex_patterns = [fixed_words_pattern] + regex_patterns
        regex_patterns_combine = r"(" + "|".join(merged_regex_patterns) + ")"
        patterns = re.compile(regex_patterns_combine, re.VERBOSE | re.UNICODE)
        recompile_regex_patterns = True
    if use_character_normalize:
        text = normalize_characters_in_text(text)
    matches = [m for m in re.finditer(patterns, text)]
    tokens = [extract_match(m) for m in matches]

    if tag:
        return tokens

    tokens = [token[0] for token in tokens]
    if use_token_normalize:
        tokens = [token_normalize(_, use_character_normalize=use_character_normalize) for _ in tokens]

    if format == "text":
        return " ".join(tokens)

    return tokens



from github import Github
import os


def upload_datasets():
    pass


def download_datasets():
    pass


def build_datasets():
    pass


if __name__ == '__main__':
    GITHUB_ACCESS_TOKEN = os.environ["GITHUB_ACCESS_TOKEN"]
    g = Github(GITHUB_ACCESS_TOKEN)
    release = g.get_repo("undertheseanlp/underthesea").get_release("open-data")
    assets = release.get_assets("abc.txt")
    print(0)



from underthesea import word_tokenize
from underthesea.models.dependency_parser import DependencyParser

uts_parser = None


def init_parser():
    global uts_parser
    if not uts_parser:
        uts_parser = DependencyParser.load('vi-dp-v1a1')
    return uts_parser


def dependency_parse(text):
    sentence = word_tokenize(text)
    parser = init_parser()
    dataset = parser.predict([sentence])
    results = dataset.sentences[0].values
    results = list(zip(results[1], results[6], results[7]))
    return results



from .token_normalize import token_normalize
from underthesea.pipeline.word_tokenize.regex_tokenize import tokenize


def text_normalize(text, tokenizer='underthesea'):
    """

    Args:
        tokenizer (str): space or underthesea
    """
    if tokenizer == 'underthesea':
        tokens = tokenize(text)
    else:
        tokens = text.split(" ")
    normalized_tokens = [token_normalize(token) for token in tokens]
    normalized_text = " ".join(normalized_tokens)
    return normalized_text



# -*- coding: utf-8 -*-

import torch
import torch.nn as nn


class ScalarMix(nn.Module):
    r"""
    Computes a parameterised scalar mixture of :math:`N` tensors, :math:`mixture = \gamma * \sum_{k}(s_k * tensor_k)`
    where :math:`s = \mathrm{softmax}(w)`, with :math:`w` and :math:`\gamma` scalar parameters.

    Args:
        n_layers (int):
            The number of layers to be mixed, i.e., :math:`N`.
        dropout (float):
            The dropout ratio of the layer weights.
            If dropout > 0, then for each scalar weight, adjust its softmax weight mass to 0
            with the dropout probability (i.e., setting the unnormalized weight to -inf).
            This effectively redistributes the dropped probability mass to all other weights.
            Default: 0.
    """

    def __init__(self, n_layers, dropout=0):
        super().__init__()

        self.n_layers = n_layers

        self.weights = nn.Parameter(torch.zeros(n_layers))
        self.gamma = nn.Parameter(torch.tensor([1.0]))
        self.dropout = nn.Dropout(dropout)

    def __repr__(self):
        s = f"n_layers={self.n_layers}"
        if self.dropout.p > 0:
            s += f", dropout={self.dropout.p}"

        return f"{self.__class__.__name__}({s})"

    def forward(self, tensors):
        r"""
        Args:
            tensors (list[~torch.Tensor]):
                :math:`N` tensors to be mixed.

        Returns:
            The mixture of :math:`N` tensors.
        """

        normed_weights = self.dropout(self.weights.softmax(-1))
        weighted_sum = sum(w * h for w, h in zip(normed_weights, tensors))

        return self.gamma * weighted_sum



import pycrfsuite
import logging
from underthesea_core import CRFFeaturizer
from underthesea.transformer.tagged_feature import lower_words
# from conlleval import evaluate_

logger = logging.getLogger(__name__)
logger.setLevel(10)
FORMAT = "%(asctime)-15s %(message)s"
logging.basicConfig(format=FORMAT)


class Trainer:
    def __init__(self, features, corpus):
        self.features = features
        self.corpus = corpus

    def train(self, params):
        transformer = CRFFeaturizer(self.features, lower_words)
        logger.info("Start feature extraction")
        X_train, y_train = transformer.transform(self.corpus.train, contain_labels=True)
        X_test, y_test = transformer.transform(self.corpus.test, contain_labels=True)
        logger.info("Finish feature extraction")

        # Train
        logger.info("Start train")
        trainer = pycrfsuite.Trainer(verbose=True)
        for xseq, yseq in zip(X_train, y_train):
            trainer.append(xseq, yseq)
        trainer.set_params(params)
        filename = 'tmp/model.tmp'
        trainer.train(filename)
        logger.info("Finish train")

        # Tagger
        logger.info("Start tagger")
        tagger = pycrfsuite.Tagger()
        tagger.open(filename)
        y_pred = [tagger.tag(x_seq) for x_seq in X_test]
        sentences = [[item[0] for item in sentence] for sentence in self.corpus.test]
        sentences = zip(sentences, y_test, y_pred)
        texts = []
        for s in sentences:
            tokens, y_true, y_pred = s
            tokens_ = ["\t".join(item) for item in zip(tokens, y_true, y_pred)]
            text = "\n".join(tokens_)
            texts.append(text)
        text = "\n\n".join(texts)
        open("tmp/output.txt", "w").write(text)
        # evaluate_("tmp/output.txt")
        logger.info("Finish tagger")



# -*- coding: utf-8 -*-

import unicodedata


def ispunct(token):
    return all(unicodedata.category(char).startswith('P')
               for char in token)


def isfullwidth(token):
    return all(unicodedata.east_asian_width(char) in ['W', 'F', 'A']
               for char in token)


def islatin(token):
    return all('LATIN' in unicodedata.name(char)
               for char in token)


def isdigit(token):
    return all('DIGIT' in unicodedata.name(char)
               for char in token)


def tohalfwidth(token):
    return unicodedata.normalize('NFKC', token)


def stripe(x, n, w, offset=(0, 0), dim=1):
    r"""
    Returns a diagonal stripe of the tensor.

    Args:
        x (~torch.Tensor): the input tensor with 2 or more dims.
        n (int): the length of the stripe.
        w (int): the width of the stripe.
        offset (tuple): the offset of the first two dims.
        dim (int): 0 if returns a horizontal stripe; 1 otherwise.

    Returns:
        a diagonal stripe of the tensor.

    Examples:
        >>> x = torch.arange(25).view(5, 5)
        >>> x
        tensor([[ 0,  1,  2,  3,  4],
                [ 5,  6,  7,  8,  9],
                [10, 11, 12, 13, 14],
                [15, 16, 17, 18, 19],
                [20, 21, 22, 23, 24]])
        >>> stripe(x, 2, 3, (1, 1))
        tensor([[ 6,  7,  8],
                [12, 13, 14]])
        >>> stripe(x, 2, 3, dim=0)
        tensor([[ 0,  5, 10],
                [ 6, 11, 16]])
    """

    x, seq_len = x.contiguous(), x.size(1)
    stride, numel = list(x.stride()), x[0, 0].numel()
    stride[0] = (seq_len + 1) * numel
    stride[1] = (1 if dim == 1 else seq_len) * numel
    return x.as_strided(size=(n, w, *x.shape[2:]),
                        stride=stride,
                        storage_offset=(offset[0] * seq_len + offset[1]) * numel)


def pad(tensors, padding_value=0, total_length=None):
    size = [len(tensors)] + [max(tensor.size(i) for tensor in tensors)
                             for i in range(len(tensors[0].size()))]
    if total_length is not None:
        assert total_length >= size[1]
        size[1] = total_length
    out_tensor = tensors[0].data.new(*size).fill_(padding_value)
    for i, tensor in enumerate(tensors):
        out_tensor[i][[slice(0, i) for i in tensor.size()]] = tensor
    return out_tensor



from sklearn.base import BaseEstimator, TransformerMixin
import string
from underthesea.pipeline.word_tokenize import tokenize
from underthesea.utils.vietnamese_features import remove_tone


negative_emoticons = {':(', '☹', '❌', '👎', '👹', '💀', '🔥', '🤔', '😏', '😐', '😑', '😒', '😓', '😔', '😕', '😖',
                      '😞', '😟', '😠', '😡', '😢', '😣', '😤', '😥', '😧', '😨', '😩', '😪', '😫', '😭', '😰', '😱',
                      '😳', '😵', '😶', '😾', '🙁', '🙏', '🚫', '>:[', ':-(', ':(', ':-c', ':c', ':-<', ':っC', ':<',
                      ':-[', ':[', ':{'}

positive_emoticons = {'=))', 'v', ';)', '^^', '<3', '☀', '☺', '♡', '♥', '✌', '✨', '❣', '❤', '🌝', '🌷', '🌸',
                      '🌺', '🌼', '🍓', '🎈', '🐅', '🐶', '🐾', '👉', '👌', '👍', '👏', '👻', '💃', '💄', '💋',
                      '💌', '💎', '💐', '💓', '💕', '💖', '💗', '💙', '💚', '💛', '💜', '💞', ':-)', ':)', ':D', ':o)',
                      ':]', ':3', ':c)', ':>', '=]', '8)'}


class Lowercase(BaseEstimator, TransformerMixin):
    def transform(self, x):
        return [s.lower() for s in x]

    def fit(self, x, y=None):
        return self


class RemoveTone(BaseEstimator, TransformerMixin):
    def remove_tone(self, s):
        return remove_tone(s)

    def transform(self, x):
        return [self.remove_tone(s) for s in x]

    def fit(self, x, y=None):
        return self


class CountEmoticons(BaseEstimator, TransformerMixin):
    def count_emoticon(self, s):
        positive_count = 0
        negative_count = 0
        for emoticon in positive_emoticons:
            positive_count += s.count(emoticon)
        for emoticon in negative_emoticons:
            negative_count += s.count(emoticon)
        return positive_count, negative_count

    def transform(self, x):
        return [self.count_emoticon(s) for s in x]

    def fit(self, x, y=None):
        return self


class Tokenize(BaseEstimator, TransformerMixin):
    def pun_num(self, s):
        for token in s.split():
            if token in string.punctuation:
                if token == '.':
                    s = s
                else:
                    s = s.replace(token, 'punc')
            else:
                s = s
        return s

    def transform(self, x):
        return [self.pun_num(tokenize(s, format='text')) for s in x]

    def fit(self, x, y=None):
        return self



from os.path import join
from pathlib import Path
import joblib
import pycrfsuite

from underthesea.transformer.tagged import TaggedTransformer


class CRFSequenceTagger:
    def __init__(self, features=None):
        self.features = features
        self.estimator = None
        self.transformer = None

    def forward(self, samples, contains_labels=False):
        if not self.transformer:
            self.transformer = TaggedTransformer(self.features)
        return self.transformer.transform(samples, contains_labels)

    def save(self, path):
        print("save features")
        joblib.dump(self.features, path)

    def load(self, base_path):
        print(base_path)
        model_path = str(Path(base_path) / "model.bin")
        estimator = pycrfsuite.Tagger()
        estimator.open(model_path)
        features = joblib.load(join(base_path, "features.bin"))
        transformer = TaggedTransformer(features)
        self.transformer = transformer
        self.estimator = estimator

    def predict(self, tokens):
        tokens = [(token, "X") for token in tokens]
        x = self.transformer.transform([tokens])[0]
        tags = self.estimator.tag(x)
        return tags



from os.path import dirname, join, exists
from torch.utils.data import Dataset
from underthesea import pos_tag
from underthesea import dependency_parse
from underthesea.utils.col_analyzer import UDAnalyzer

BOT_VERSION = "underthesea.v1.3.2"
PROJECT_FOLDER = dirname(dirname(dirname(__file__)))
DATASETS_FOLDER = join(PROJECT_FOLDER, "datasets")
COL_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL")


class UDSentence:
    def __init__(self, rows, headers=None):
        self.rows = rows
        self.key_orders = ["doc_url", "date", "sent_id", "type", "authors", "text"]
        self.headers = headers

    def __str__(self):
        content = ""
        for key in self.key_orders:
            value = self.headers[key]
            content += f"# {key} = {value}\n"
        content += self.get_ud_str()
        return content

    def get_ud_str(self):
        return "\n".join(["\t".join(row) for row in self.rows])

    @staticmethod
    def _extract_header(content):
        content = content[2:].strip()
        index = content.find("=")
        key = content[:index].strip()
        value = content[index + 2:].strip()
        return [key, value]

    @staticmethod
    def load(content):
        data = content.split("\n")
        headers = [row for row in data if row.startswith("# ")]
        headers = dict([UDSentence._extract_header(content) for content in headers])
        rows = [row for row in data if not row.startswith("# ")]
        rows = [r.split("\t") for r in rows]
        s = UDSentence(rows=rows, headers=headers)
        return s

    @staticmethod
    def load_from_raw_content(raw_content):
        sentences = raw_content.split("\n")
        headers = sentences[:3]
        headers = dict([UDSentence._extract_header(_) for _ in headers])
        text = sentences[-1]
        headers["text"] = text
        headers["type"] = "bronze"
        headers["authors"] = BOT_VERSION
        pos_tags = pos_tag(text)
        dp_tags = dependency_parse(text)
        rows = [(item[0][0], item[0][1], str(item[1][1]), item[1][2]) for item in zip(pos_tags, dp_tags)]
        s = UDSentence(rows=rows, headers=headers)
        return s

    @staticmethod
    def load_from_raw_text(text):
        raw_content = ''
        raw_content += "# doc_url = DOC_URL\n"
        raw_content += "# date = DATE\n"
        raw_content += "# sent_id = 0\n"
        raw_content += text.strip()
        return UDSentence.load_from_raw_content(raw_content)


class UDDataset(Dataset):

    def __init__(self, sentences):
        super().__init__()
        self.sentences = sentences
        self.generate_indices()

    def generate_indices(self):
        self.sent_id_sent_index_map = dict([[s.headers["sent_id"], key] for key, s in enumerate(self.sentences)])

    def get_by_sent_id(self, sent_id):
        if sent_id in self.sent_id_sent_index_map:
            index = self.sent_id_sent_index_map[sent_id]
            return self.sentences[index]
        return None

    def __getitem__(self, index):
        return self.sentences[index]

    def __len__(self):
        return len(self.sentences)

    @staticmethod
    def load(ud_file):
        sentences = open(ud_file).read().split("\n\n")
        sentences = [UDSentence.load(s) for s in sentences]
        dataset = UDDataset(sentences)
        return dataset

    @staticmethod
    def load_from_raw_file(raw_file):
        with open(raw_file) as f:
            rows = f.read().strip().split("\n\n")
        sentences = [UDSentence.load_from_raw_content(content) for content in rows]
        dataset = UDDataset(sentences)
        return dataset

    def merge_sentence(self, s1, dataset):
        sent_id = s1.headers['sent_id']
        target_sentence = dataset.get_by_sent_id(sent_id)
        if target_sentence is None:
            return s1
        if target_sentence.headers['type'] == 'silver':
            return target_sentence
        return s1

    def merge(self, dataset):
        self.sentences = [self.merge_sentence(s, dataset) for s in self.sentences]

    def write(self, target_file):
        content = "\n\n".join([str(s) for s in self.sentences])
        with open(target_file, "w") as f:
            f.write(content)


if __name__ == '__main__':
    file = "202108.txt"  # lyrics.txt or 202108.txt
    raw_file = join(COL_FOLDER, "corpus", "raw", file)
    generated_dataset = UDDataset.load_from_raw_file(raw_file)

    current_file = join(COL_FOLDER, "corpus", "ud", file)
    if exists(current_file):
        current_dataset = UDDataset.load(current_file)

        generated_dataset.merge(current_dataset)

    target_file = join(COL_FOLDER, "corpus", "ud", file)
    generated_dataset.write(target_file)

    analyzer = UDAnalyzer()
    analyzer.analyze(generated_dataset)
    analyzer.analyze_today_words(generated_dataset)



from underthesea.corpus.corpus import Corpus
from underthesea.corpus.data import Document
from underthesea.feature_engineering.unicode import UnicodeTransformer
from underthesea.corpus.plaintext_corpus import PlainTextCorpus
from underthesea.corpus.readers.dictionary_loader import DictionaryLoader

viet_dict_11K = DictionaryLoader('Viet11K.txt')
viet_dict_22K = DictionaryLoader('Viet22K.txt')
viet_dict_39K = DictionaryLoader('Viet39K.txt')
viet_dict_74K = DictionaryLoader('Viet74K.txt')

__all__ = ['Corpus', 'Document', 'UnicodeTransformer', 'PlainTextCorpus', 'DictionaryLoader']



# -*- coding: utf-8 -*-
from underthesea import word_tokenize
from underthesea.models.fast_crf_sequence_tagger import FastCRFSequenceTagger
from .model_crf import CRFPOSTagPredictor
from os.path import dirname, join

pos_model_v2 = None


def pos_tag(sentence, format=None, model=None):
    global pos_model_v2
    """
    Vietnamese POS tagging

    Parameters
    ==========

    sentence: {unicode, str}
        Raw sentence

    Returns
    =======
    tokens: list of tuple with word, pos tag
        tagged sentence
    Examples
    --------
    >>> # -*- coding: utf-8 -*-
    >>> from underthesea import pos_tag
    >>> sentence = "Chợ thịt chó nổi tiếng ở TPHCM bị truy quét"
    >>> pos_tag(sentence)
    [('Chợ', 'N'),
    ('thịt', 'N'),
    ('chó', 'N'),
    ('nổi tiếng', 'A'),
    ('ở', 'E'),
    ('TPHCM', 'Np'),
    ('bị', 'V'),
    ('truy quét', 'V')]
    """
    sentence = word_tokenize(sentence)
    if model == "v2.0":
        if pos_model_v2 is None:
            pos_model_v2 = FastCRFSequenceTagger()
            wd = dirname(__file__)
            pos_model_v2.load(join(wd, "models", "pos_crf_vlsp2013_20230303"))
            tokens = sentence
            features = [[token] for token in sentence]
            tags = pos_model_v2.predict(features)
            # output of pos_model_v2 in in BOI format B-N, B-CH, B-V,...
            # remove prefix B-
            tags = [tag[2:] for tag in tags]
            result = list(zip(tokens, tags))
    else:
        crf_model = CRFPOSTagPredictor.Instance()
        result = crf_model.predict(sentence, format)
    return result



# -*- coding: utf-8 -*-

import torch.nn as nn
from torch.nn.utils.rnn import pad_sequence
from transformers import AutoConfig, AutoModel

from underthesea.modules.scalar_mix import ScalarMix


class BertEmbedding(nn.Module):
    r"""
    A module that directly utilizes the pretrained models in `transformers`_ to produce BERT representations.

    While mainly tailored to provide input preparation and post-processing for the BERT model,
    it is also compatiable with other pretrained language models like XLNet, RoBERTa and ELECTRA, etc.

    Args:
        model (str):
            Path or name of the pretrained models registered in `transformers`_, e.g., ``'bert-base-cased'``.
        n_layers (int):
            The number of layers from the model to use.
            If 0, uses all layers.
        n_out (int):
            The requested size of the embeddings.
            If 0, uses the size of the pretrained embedding model.
        pad_index (int):
            The index of the padding token in the BERT vocabulary. Default: 0.
        max_len (int):
            Sequences should not exceed the specfied max length. Default: 512.
        dropout (float):
            The dropout ratio of BERT layers. Default: 0.
            This value will be passed into the :class:`ScalarMix` layer.
        requires_grad (bool):
            If ``True``, the model parameters will be updated together with the downstream task.
            Default: ``False``.

    .. _transformers:
        https://github.com/huggingface/transformers
    """

    def __init__(self, model, n_layers, n_out, pad_index=0, max_len=512, dropout=0, requires_grad=False):
        super().__init__()

        self.model = model
        self.bert = AutoModel.from_pretrained(model,
                                              config=AutoConfig.from_pretrained(model, output_hidden_states=True))
        self.bert = self.bert.requires_grad_(requires_grad)
        self.n_layers = n_layers or self.bert.config.num_hidden_layers
        self.hidden_size = self.bert.config.hidden_size
        self.n_out = n_out or self.hidden_size
        self.pad_index = pad_index
        self.max_len = max_len
        self.dropout = dropout
        self.requires_grad = requires_grad

        self.scalar_mix = ScalarMix(self.n_layers, dropout)
        self.projection = nn.Linear(self.hidden_size, self.n_out, False) if self.hidden_size != n_out else nn.Identity()

    def __repr__(self):
        s = f"{self.model}, n_layers={self.n_layers}, n_out={self.n_out}, pad_index={self.pad_index}"
        if self.max_len is not None:
            s += f", max_len={self.max_len}"
        if self.dropout > 0:
            s += f", dropout={self.dropout}"
        if self.requires_grad:
            s += f", requires_grad={self.requires_grad}"

        return f"{self.__class__.__name__}({s})"

    def forward(self, subwords):
        r"""
        Args:
            subwords (~torch.Tensor): ``[batch_size, seq_len, fix_len]``.

        Returns:
            ~torch.Tensor:
                BERT embeddings of shape ``[batch_size, seq_len, n_out]``.
        """
        batch_size, seq_len, fix_len = subwords.shape
        if self.max_len and seq_len > self.max_len:
            raise RuntimeError(f"Token indices sequence length is longer than the specified max length "
                               f"({seq_len} > {self.max_len})")

        mask = subwords.ne(self.pad_index)
        lens = mask.sum((1, 2))
        # [batch_size, n_subwords]
        subwords = pad_sequence(subwords[mask].split(lens.tolist()), True)
        bert_mask = pad_sequence(mask[mask].split(lens.tolist()), True)
        # return the hidden states of all layers
        bert = self.bert(subwords, attention_mask=bert_mask.float())[-1]
        # [n_layers, batch_size, n_subwords, hidden_size]
        bert = bert[-self.n_layers:]
        # [batch_size, n_subwords, hidden_size]
        bert = self.scalar_mix(bert)
        # [batch_size, n_subwords]
        bert_lens = mask.sum(-1)
        bert_lens = bert_lens.masked_fill_(bert_lens.eq(0), 1)
        # [batch_size, seq_len, fix_len, hidden_size]
        embed = bert.new_zeros(*mask.shape, self.hidden_size)
        embed = embed.masked_scatter_(mask.unsqueeze(-1), bert[bert_mask])
        # [batch_size, seq_len, hidden_size]
        embed = embed.sum(2) / bert_lens.unsqueeze(-1)
        embed = self.projection(embed)

        return embed



import argparse
import json
import os
import pickle

import numpy as np
import torch

from .config import FLAGS
from .torch_model import Generator


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self


def load_checkpoint(filepath, device):
    assert os.path.isfile(filepath)
    print("Loading '{}'".format(filepath))
    checkpoint_dict = torch.load(filepath, map_location=device)
    print("Complete.")
    return checkpoint_dict


def convert_to_haiku(a, h, device):
    generator = Generator(h).to(device)
    state_dict_g = load_checkpoint(a.checkpoint_file, device)
    generator.load_state_dict(state_dict_g["generator"])
    generator.eval()
    generator.remove_weight_norm()
    hk_map = {}
    for a, b in generator.state_dict().items():
        print(a, b.shape)
        if a.startswith("conv_pre"):
            a = "generator/~/conv1_d"
        elif a.startswith("conv_post"):
            a = "generator/~/conv1_d_1"
        elif a.startswith("ups."):
            ii = a.split(".")[1]
            a = f"generator/~/ups_{ii}"
        elif a.startswith("resblocks."):
            _, x, y, z, _ = a.split(".")
            a = f"generator/~/res_block1_{x}/~/{y}_{z}"
        print(a, b.shape)
        if a not in hk_map:
            hk_map[a] = {}
        if len(b.shape) == 1:
            hk_map[a]["b"] = b.numpy()
        else:
            if "ups" in a:
                hk_map[a]["w"] = np.rot90(b.numpy(), k=1, axes=(0, 2))
            elif "conv" in a:
                hk_map[a]["w"] = np.swapaxes(b.numpy(), 0, 2)
            else:
                hk_map[a]["w"] = b.numpy()

    FLAGS.ckpt_dir.mkdir(parents=True, exist_ok=True)
    with open(FLAGS.ckpt_dir / "hk_hifi.pickle", "wb") as f:
        pickle.dump(hk_map, f)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--checkpoint-file", required=True)
    parser.add_argument("--config-file", required=True)
    a = parser.parse_args()

    config_file = a.config_file
    with open(config_file) as f:
        data = f.read()

    json_config = json.loads(data)
    h = AttrDict(json_config)

    device = torch.device("cpu")
    convert_to_haiku(a, h, device)


if __name__ == "__main__":
    main()



from underthesea.file_utils import MODELS_FOLDER
from underthesea.pipeline.say.viettts_.hifigan.mel2wave import mel2wave
from underthesea.pipeline.say.viettts_.nat.text2mel import text2mel
from underthesea.pipeline.say.viettts_ import nat_normalize_text
import numpy as np
from playsound import playsound
import time
from os.path import join


model_path = join(MODELS_FOLDER, "VIET_TTS_V0_4_1")


def text_to_speech(text):
    # prevent too long text
    if len(text) > 500:
        text = text[:500]
    text = nat_normalize_text(text)
    mel = text2mel(
        text,
        join(model_path, "lexicon.txt"),
        0.2,
        join(model_path, "acoustic_latest_ckpt.pickle"),
        join(model_path, "duration_latest_ckpt.pickle"),
    )
    wave = mel2wave(mel, join(model_path, "config.json"), join(model_path, "hk_hifi.pickle"))
    return (wave * (2**15)).astype(np.int16)


def say(text, outfile="sound.wav", play=False):
    y = text_to_speech(text)
    # write y array to sound.wav
    import soundfile as sf
    sf.write(outfile, y, 16_000)
    time_str = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
    print(f"\n{time_str}: {text}")
    if play:
        playsound(outfile)
    return 16_000, y


def think_and_say(text, outfile="sound.wav"):
    # create two processes, one for thinking, one for saying
    import multiprocessing
    p1 = multiprocessing.Process(target=terminal_thinking)
    p2 = multiprocessing.Process(target=say, args=(text, outfile))
    p1.start()
    p2.start()
    # while p2 is end, terminate p1
    p2.join()
    p1.terminate()
    playsound(outfile)


def terminal_thinking():
    # each 300ms, print a dot
    import time
    while True:
        print(".", end="")
        time.sleep(0.5)


if __name__ == '__main__':
    print(time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()))
    think_and_say("xin chào")
    think_and_say("Có đấy. Tình yêu giống như một ly kem, nếu không ăn nó đúng lúc, nó sẽ tan chảy. Haha, anh thấy thế nào?")



import os
from os import listdir
from os.path import join
from underthesea import sent_tokenize
from multiprocessing import Pool

from underthesea.file_utils import UNDERTHESEA_FOLDER

WIKI_FOLDER = join(UNDERTHESEA_FOLDER, "data", "viwiki-20210720")
RAW_FOLDER = join(WIKI_FOLDER, "text", "AA")
CLEANED_FOLDER = join(WIKI_FOLDER, "cleaned", "AA")

os.makedirs(CLEANED_FOLDER, exist_ok=True)


def check_line(line):
    if len(line) < 30:
        return False
    if line.startswith("<"):
        return False
    return True


def clean(file):
    print(file)
    with open(join(RAW_FOLDER, file)) as f:
        content = f.read()
    with open(join(CLEANED_FOLDER, file), 'w') as out_file:
        out_file.write('')
    with open(join(CLEANED_FOLDER, file), 'a') as out_file:
        for line in content.split("\n"):
            if not check_line(line):
                continue
            sents = sent_tokenize(line)
            for sent in sents:
                if not check_line(sent):
                    continue
                out_file.write(sent.strip() + '\n')
    return


if __name__ == '__main__':
    files = sorted(listdir(RAW_FOLDER))
    with Pool(10) as p:
        p.map(clean, files)



import logging.config

logging.config.dictConfig(
    {
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {"standard": {"format": "%(asctime)-15s %(message)s"}},
        "handlers": {
            "console": {
                "level": "INFO",
                "class": "logging.StreamHandler",
                "formatter": "standard",
                "stream": "ext://sys.stdout",
            }
        },
        "loggers": {
            "underthesea": {"handlers": ["console"], "level": "INFO", "propagate": False}
        },
    }
)

logger = logging.getLogger("underthesea")


###########################################################
# Print Table with orgtbl (replace tabulate)
###########################################################
def generate_table(data, headers=None):
    content = ''
    # calculate max_widths
    num_columns = len(data[0])
    max_widths = [0] * len(data[0])
    for item in data:
        for i in range(num_columns):
            max_widths[i] = max(max_widths[i], len(str(item[i])))
    # generate table headers
    if headers:
        for i, item in enumerate(headers):
            max_widths[i] = max(max_widths[i], len(str(item)))
        header_str = ''
        for i, item in enumerate(headers):
            header_str += '| {item:{width}} '.format(item=item, width=max_widths[i])
        header_str += '|\n'
        for i, item in enumerate(headers):
            border_str = '|' if i == 0 else '+'
            header_str += border_str + '-' * (max_widths[i] + 2)
        header_str += '|\n'
    content += header_str
    # generate table body
    for row in data:
        data_str = ''
        for i, item in enumerate(row):
            data_str += '| {item:{width}} '.format(item=str(item), width=max_widths[i])
        data_str += '|\n'
        content += data_str
    return content


def print_table(data, headers=None):
    content = generate_table(data, headers)
    print(content)



from underthesea.corpus import Corpus, Document
from os.path import join
from os import listdir, mkdir

from underthesea.feature_engineering.unicode import UnicodeTransformer
from underthesea.util.file_io import write
import io


class PlainTextCorpus(Corpus):
    """class for handling plain text corpus
    """

    def __init__(self):
        self.documents = None

    def load(self, folder):
        """load plaintext folder to documents and apply unicode transformer

        :param str folder: path to directory
        :type folder: string
        """
        ids = listdir(folder)
        files = [join(folder, f) for f in ids]
        contents = []
        for file in files:
            with io.open(file, "r", encoding="utf-8") as f:
                contents.append(f.read())
        documents = []

        for id, content in zip(ids, contents):
            document = Document(id)
            sentences = content.split("\n")
            unicode_transformer = UnicodeTransformer()
            sentences = [unicode_transformer.transform(sentence) for sentence in sentences]
            document.set_sentences(sentences)
            documents.append(document)
        self.documents = documents

    def save(self, folder):
        """save corpus to files

        :param str folder: path to directory
        :type folder: string
        """
        try:
            mkdir(folder)
        except Exception:
            pass
        for document in self.documents:
            filename = join(folder, document.id)
            content = u"\n".join(document.sentences)
            write(filename, content)



# -*- coding: utf-8 -*-
from underthesea import chunk
from .model_crf import CRFNERPredictor


def ner(sentence, format=None, deep=False):
    """
    Location and classify named entities in text

    Parameters
    ==========

    sentence: {unicode, str}
        raw sentence

    Returns
    =======
    tokens: list of tuple with word, pos tag, chunking tag, ner tag tagged sentence

    Examples
    --------

    >>> # -*- coding: utf-8 -*-
    >>> from underthesea import ner
    >>> sentence = "Ông Putin ca ngợi những thành tựu vĩ đại của Liên Xô"
    >>> ner(sentence)
    [('Ông', 'Nc', 'B-NP', 'O'),
    ('Putin', 'Np', 'B-NP', 'B-PER'),
    ('ca ngợi', 'V', 'B-VP', 'O'),
    ('những', 'L', 'B-NP', 'O'),
    ('thành tựu', 'N', 'B-NP', 'O'),
    ('vĩ đại', 'A', 'B-AP', 'O'),
    ('của', 'E', 'B-PP', 'O'),
    ('Liên Xô', 'Np', 'B-NP', 'B-LOC')]
    """
    if not deep:
        sentence = chunk(sentence)
        crf_model = CRFNERPredictor.Instance()
        result = crf_model.predict(sentence, format)
        return result
    else:
        from .model_transformers import nlp
        output = nlp(sentence)
        if len(output) == 0:
            return []
        entities = [output[0]]
        for item in output[1:]:
            if item["word"].startswith("##"):
                entities[-1]["word"] = entities[-1]["word"] + item["word"][2:]
                entities[-1]["end"] = item["end"]
            else:
                entities.append(item)
        return entities



import sys


class NumberRemover:
    """
    Remove numbers in documents
    """

    def __init__(self):
        pass

    def _remove(self, document):
        numbers = "0123456789"

        if sys.version_info < (3, 0, 0):
            return document.translate(None, numbers)
        else:
            remover = str.maketrans("", "", numbers)
            return document.translate(remover)

    def transform(self, raw_documents):
        """
        Remove number in each document

        Parameters
        ----------
        raw_documents : iterable
            An iterable which yields either str, unicode

        Returns
        -------
        X : iterable
            cleaned documents
        """
        return [self._remove(document) for document in raw_documents]



# -*- coding: utf-8 -*-

from collections import defaultdict
from collections.abc import Iterable


class Vocab(object):
    r"""
    Defines a vocabulary object that will be used to numericalize a field.

    Args:
        counter (~collections.Counter):
            :class:`~collections.Counter` object holding the frequencies of each value found in the data.
        min_freq (int):
            The minimum frequency needed to include a token in the vocabulary. Default: 1.
        specials (list[str]):
            The list of special tokens (e.g., pad, unk, bos and eos) that will be prepended to the vocabulary. Default: [].
        unk_index (int):
            The index of unk token. Default: 0.

    Attributes:
        itos:
            A list of token strings indexed by their numerical identifiers.
        stoi:
            A :class:`~collections.defaultdict` object mapping token strings to numerical identifiers.
    """

    def __init__(self, counter, min_freq=1, specials=[], unk_index=0):
        self.itos = list(specials)
        self.stoi = defaultdict(lambda: unk_index)
        self.stoi.update({token: i for i, token in enumerate(self.itos)})
        self.extend([token for token, freq in counter.items()
                     if freq >= min_freq])
        self.unk_index = unk_index
        self.n_init = len(self)

    def __len__(self):
        return len(self.itos)

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.stoi[key]
        elif not isinstance(key, Iterable):
            return self.itos[key]
        elif isinstance(key[0], str):
            return [self.stoi[i] for i in key]
        else:
            return [self.itos[i] for i in key]

    def __contains__(self, token):
        return token in self.stoi

    def __getstate__(self):
        # avoid picking defaultdict
        attrs = dict(self.__dict__)
        # cast to regular dict
        attrs['stoi'] = dict(self.stoi)
        return attrs

    def __setstate__(self, state):
        stoi = defaultdict(lambda: self.unk_index)
        stoi.update(state['stoi'])
        state['stoi'] = stoi
        self.__dict__.update(state)

    def extend(self, tokens):
        self.itos.extend(sorted(set(tokens).difference(self.stoi)))
        self.stoi.update({token: i for i, token in enumerate(self.itos)})



from os import makedirs
from os.path import join, exists
from shutil import rmtree
from seqeval.metrics import classification_report, accuracy_score
from underthesea.file_utils import UNDERTHESEA_FOLDER
from underthesea.transformer.tagged import TaggedTransformer
import pycrfsuite

from underthesea.utils import logger


class ModelTrainer:
    def __init__(self, tagger, corpus):
        self.tagger = tagger
        self.corpus = corpus

    def train(self, base_path, params):
        base_path = join(UNDERTHESEA_FOLDER, base_path)
        if exists(base_path):
            rmtree(base_path)
        makedirs(base_path)
        features = self.tagger.features
        print(features)
        transformer = TaggedTransformer(features)
        logger.info("Start feature extraction")
        train_samples = self.corpus.train
        test_samples = self.corpus.test
        X_train, y_train = transformer.transform(train_samples, contain_labels=True)
        X_test, y_test = transformer.transform(test_samples, contain_labels=True)
        logger.info("Finish feature extraction")

        # Train
        logger.info("Start train")
        trainer = pycrfsuite.Trainer(verbose=True)
        for xseq, yseq in zip(X_train, y_train):
            trainer.append(xseq, yseq)
        trainer.set_params(params)
        filename = join(base_path, 'model.bin')
        trainer.train(filename)
        logger.info("Finish train")

        # Tagger
        logger.info("Start tagger")
        tagger = pycrfsuite.Tagger()
        tagger.open(filename)
        y_pred = [tagger.tag(x_seq) for x_seq in X_test]
        y_true = y_test
        print(classification_report(y_true, y_pred, digits=4))
        print(f'Accuracy: {accuracy_score(y_true, y_pred):.4f}')
        sentences = [[item[0] for item in sentence] for sentence in self.corpus.test]
        sentences = zip(sentences, y_test, y_pred)
        texts = []
        for s in sentences:
            tokens, y_true, y_pred = s
            tokens_ = ["\t".join(item) for item in zip(tokens, y_true, y_pred)]
            text = "\n".join(tokens_)
            texts.append(text)
        text = "\n\n".join(texts)
        open(join(base_path, "output.txt"), "w").write(text)

        with open(join(base_path, "model.metadata"), "w") as f:
            f.write("model: CRFSequenceTagger")
        self.tagger.save(join(base_path, "features.bin"))

        logger.info("Finish tagger")



from underthesea.corpus import Corpus


class CONLLCorpus(Corpus):
    def __init__(self, train, dev=None, test=None, name: str = 'corpus'):
        self.name: str = name
        self._train = train
        self._dev = dev
        self._test = test

    @property
    def train(self):
        return self._train

    @property
    def dev(self):
        return self._dev

    @property
    def test(self):
        return self._test



# -*- coding: utf-8 -*-


# References: https://github.com/flairNLP/flair/blob/master/flair/data.py#L1049
class Corpus:
    """Corpus is fundamental resource of NLP
    """

    def __init__(self):
        self._train = None
        self._dev = None
        self._test = None

    def load(self, folder):
        pass

    def save(self, folder):
        pass

    @property
    def train(self):
        return self._train

    @property
    def dev(self):
        return self._dev

    @property
    def test(self):
        return self._test

    def downsample(self, percentage: float = 0.1, downsample_train=True, downsample_dev=True, downsample_test=True):
        """ TODO: implement this
        Ref: Flair

        :param percentage:
        :param downsample_train:
        :param downsample_dev:
        :param downsample_test:
        :return:
        """
        if downsample_train:
            self._train = self._downsample_to_proportion(self.train, percentage)

        if downsample_dev:
            self._dev = self._downsample_to_proportion(self.dev, percentage)

        if downsample_test:
            self._test = self._downsample_to_proportion(self.test, percentage)

        return self



# flake8: noqa

import pickle
from functools import partial
from typing import Deque

import haiku as hk
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import optax
from tqdm.auto import tqdm
from vietTTS.nat.config import AcousticInput

from .config import FLAGS, AcousticInput
from .data_loader import load_textgrid_wav
from .dsp import MelFilter
from .model import AcousticModel
from .utils import print_flags


@hk.transform_with_state
def net(x):
    return AcousticModel(is_training=True)(x)


@hk.transform_with_state
def val_net(x):
    return AcousticModel(is_training=False)(x)


def loss_fn(params, aux, rng, inputs: AcousticInput, is_training=True):
    melfilter = MelFilter(
        FLAGS.sample_rate, FLAGS.n_fft, FLAGS.mel_dim, FLAGS.fmin, FLAGS.fmax
    )
    mels = melfilter(inputs.wavs.astype(jnp.float32) / (2 ** 15))
    B, L, D = mels.shape
    go_frame = jnp.zeros((B, 1, D), dtype=jnp.float32)
    inp_mels = jnp.concatenate((go_frame, mels[:, :-1, :]), axis=1)
    n_frames = inputs.durations * FLAGS.sample_rate / (FLAGS.n_fft // 4)
    inputs = inputs._replace(mels=inp_mels, durations=n_frames)
    model = net if is_training else val_net
    (mel1_hat, mel2_hat), new_aux = model.apply(params, aux, rng, inputs)
    loss1 = (jnp.square(mel1_hat - mels) + jnp.square(mel2_hat - mels)) / 2
    loss2 = (jnp.abs(mel1_hat - mels) + jnp.abs(mel2_hat - mels)) / 2
    loss = jnp.mean((loss1 + loss2) / 2, axis=-1)
    num_frames = (inputs.wav_lengths // (FLAGS.n_fft // 4))[:, None]
    mask = jnp.arange(0, L)[None, :] < num_frames
    loss = jnp.sum(loss * mask) / jnp.sum(mask)
    return (loss, new_aux) if is_training else (loss, new_aux, mel2_hat, mels)


train_loss_fn = partial(loss_fn, is_training=True)
val_loss_fn = jax.jit(partial(loss_fn, is_training=False))

loss_vag = jax.value_and_grad(train_loss_fn, has_aux=True)


def initial_state(optimizer, batch):
    rng = jax.random.PRNGKey(42)
    params, aux = hk.transform_with_state(lambda x: AcousticModel(True)(x)).init(
        rng, batch
    )
    optim_state = optimizer.init(params)
    return params, aux, rng, optim_state


def train():

    optimizer = optax.chain(
        optax.clip_by_global_norm(1.0),
        optax.adamw(FLAGS.learning_rate, weight_decay=FLAGS.weight_decay),
    )

    @jax.jit
    def update(params, aux, rng, optim_state, inputs):
        rng, new_rng = jax.random.split(rng)
        (loss, new_aux), grads = loss_vag(params, aux, rng, inputs)
        updates, new_optim_state = optimizer.update(grads, optim_state, params)
        new_params = optax.apply_updates(updates, params)
        return loss, (new_params, new_aux, new_rng, new_optim_state)

    train_data_iter = load_textgrid_wav(
        FLAGS.data_dir,
        FLAGS.max_phoneme_seq_len,
        FLAGS.batch_size,
        FLAGS.max_wave_len,
        "train",
    )
    val_data_iter = load_textgrid_wav(
        FLAGS.data_dir,
        FLAGS.max_phoneme_seq_len,
        FLAGS.batch_size,
        FLAGS.max_wave_len,
        "val",
    )
    melfilter = MelFilter(
        FLAGS.sample_rate, FLAGS.n_fft, FLAGS.mel_dim, FLAGS.fmin, FLAGS.fmax
    )
    batch = next(train_data_iter)
    batch = batch._replace(mels=melfilter(batch.wavs.astype(jnp.float32) / (2 ** 15)))
    params, aux, rng, optim_state = initial_state(optimizer, batch)
    losses = Deque(maxlen=1000)
    val_losses = Deque(maxlen=100)

    last_step = -1

    # loading latest checkpoint
    ckpt_fn = FLAGS.ckpt_dir / "acoustic_latest_ckpt.pickle"
    if ckpt_fn.exists():
        print("Resuming from latest checkpoint at", ckpt_fn)
        with open(ckpt_fn, "rb") as f:
            dic = pickle.load(f)
            last_step, params, aux, rng, optim_state = (
                dic["step"],
                dic["params"],
                dic["aux"],
                dic["rng"],
                dic["optim_state"],
            )

    tr = tqdm(
        range(last_step + 1, FLAGS.num_training_steps + 1),
        desc="training",
        total=FLAGS.num_training_steps + 1,
        initial=last_step + 1,
    )
    for step in tr:
        batch = next(train_data_iter)
        loss, (params, aux, rng, optim_state) = update(
            params, aux, rng, optim_state, batch
        )
        losses.append(loss)

        if step % 10 == 0:
            val_batch = next(val_data_iter)
            val_loss, val_aux, predicted_mel, gt_mel = val_loss_fn(
                params, aux, rng, val_batch
            )
            val_losses.append(val_loss)
            attn = jax.device_get(val_aux["acoustic_model"]["attn"])
            predicted_mel = jax.device_get(predicted_mel[0])
            gt_mel = jax.device_get(gt_mel[0])

        if step % 1000 == 0:
            loss = sum(losses).item() / len(losses)
            val_loss = sum(val_losses).item() / len(val_losses)
            tr.write(f"step {step}  train loss {loss:.3f}  val loss {val_loss:.3f}")

            # saving predicted mels
            plt.figure(figsize=(10, 10))
            plt.subplot(3, 1, 1)
            plt.imshow(predicted_mel.T, origin="lower", aspect="auto")
            plt.subplot(3, 1, 2)
            plt.imshow(gt_mel.T, origin="lower", aspect="auto")
            plt.subplot(3, 1, 3)
            plt.imshow(attn.T, origin="lower", aspect="auto")
            plt.tight_layout()
            plt.savefig(FLAGS.ckpt_dir / f"mel_{step}.png")
            plt.close()

            # saving checkpoint
            with open(ckpt_fn, "wb") as f:
                pickle.dump(
                    {
                        "step": step,
                        "params": params,
                        "aux": aux,
                        "rng": rng,
                        "optim_state": optim_state,
                    },
                    f,
                )


if __name__ == "__main__":
    print_flags(FLAGS.__dict__)
    if not FLAGS.ckpt_dir.exists():
        print("Create checkpoint dir at", FLAGS.ckpt_dir)
        FLAGS.ckpt_dir.mkdir(parents=True, exist_ok=True)
    train()



import sys
from underthesea.datasets.vlsp2013_wtk.revise_1 import revise_corpus_vlsp2013_wtk

SUPPORTED_CORPUS = set(["VLSP2013-WTK"])


def revise_corpus(corpus_name):
    if corpus_name not in SUPPORTED_CORPUS:
        print(f"Corpus {corpus_name} is not supported")
        sys.exit(1)
    if corpus_name == "VLSP2013-WTK":
        revise_corpus_vlsp2013_wtk()



def preprocess_word_tokenize_dataset_split(data):
    tag_values = data.features['tags'].feature.names
    output = []
    for item in data:
        s = []
        tokens = item['tokens']
        tags = item['tags']
        for token, tag in zip(tokens, tags):
            s.append([token, tag_values[tag]])
        output.append(s)
    return output


def preprocess_word_tokenize_dataset(dataset):
    data = {}
    data["train"] = preprocess_word_tokenize_dataset_split(dataset["train"])
    data["validation"] = preprocess_word_tokenize_dataset_split(dataset["validation"])
    data["test"] = preprocess_word_tokenize_dataset_split(dataset["test"])
    return data



# -*- coding: utf-8 -*-
import os
import click
import platform

from underthesea.datasets.vlsp2013_wtk.revise_corpus import revise_corpus
from underthesea.corpus.validate_corpus import validate_corpus, DEFAULT_MAX_ERROR
from underthesea.data_fetcher import DataFetcher
from underthesea.model_fetcher import ModelFetcher


@click.group()
def main(args=None):
    """Console script for underthesea"""
    pass


@main.command()
@click.option('-a', '--all', is_flag=True, required=False)
def list_model(all):
    ModelFetcher.list(all)


@main.command()
@click.argument('model', required=True)
def download_model(model):
    ModelFetcher.download(model)


@main.command()
@click.argument('text', required=True)
def say(text):
    from underthesea.pipeline.say import say as underthesea_say
    underthesea_say(text, play=True)


@main.command()
@click.argument('model', required=True)
def remove_model(model):
    ModelFetcher.remove(model)


@main.command()
@click.option('-a', '--all', is_flag=True, required=False)
def list_data(all):
    DataFetcher.list(all)


@main.command()
@click.argument('dataset', required=True)
@click.argument('url', required=False)
def download_data(dataset, url):
    DataFetcher.download_data(dataset, url)


@main.command()
@click.argument('data', required=True)
def remove_data(data):
    DataFetcher.remove(data)


@main.command()
@click.option('-t', '--type', required=True)
@click.option('-c', '--corpus', required=True)
@click.option('--max-error', default=DEFAULT_MAX_ERROR, type=int)
def validate(type, corpus, max_error):
    validate_corpus(type, corpus, max_error)


@main.command()
@click.option('-c', '--corpus', required=True)
def revise(corpus):
    revise_corpus(corpus)


@main.command()
def info():
    version_file = os.path.join(os.path.dirname(__file__), 'VERSION')
    underthesea_version = open(version_file).read().strip()
    python_version = platform.python_version()
    system_info = f"{platform.system()}{platform.release()}"
    print("")
    print("ENVIRONMENT")
    print(f" underthesea version : {underthesea_version}")
    print(f"      python version : {python_version}")
    print(f"  system information : {system_info}")

    print("")
    print("MODULES")
    print("       sent_tokenize : OK")
    print("      text_normalize : OK")
    print("       word_tokenize : OK")
    print("             pos_tag : OK")
    print("               chunk : OK")
    print("                 ner : OK")
    print("            classify : SUSPENDED")
    print("           sentiment : SUSPENDED")
    print("    dependency_parse : SUSPENDED")
    print("         lang_detect : OK")
    print("           resources : OK")


if __name__ == "__main__":
    main()



import fasttext
import os
from underthesea.model_fetcher import ModelFetcher

fasttext.FastText.eprint = lambda x: None
lang_detect_model = None


def lang_detect(text):
    global lang_detect_model
    model_name = "LANG_DETECT_FAST_TEXT"
    model_path = ModelFetcher.get_model_path(model_name)
    if not lang_detect_model:
        if not os.path.exists(model_path):
            ModelFetcher.download(model_name)
        try:
            lang_detect_model = fasttext.load_model(str(model_path))
        except Exception:
            pass

    predictions = lang_detect_model.predict(text)
    language = predictions[0][0].replace('__label__', '')
    return language



from sklearn import feature_extraction
import numpy as np

from underthesea.transformer.count import CountVectorizer


class TfidfVectorizer(feature_extraction.text.TfidfVectorizer):
    """
    Convert a collection of raw documents to a matrix of TF-IDF features.

    Parameters
    ----------
    input : string {'filename', 'file', 'content'}
        If 'filename', the sequence passed as an argument to fit is
        expected to be a list of filenames that need reading to fetch
        the raw content to analyze.
        If 'file', the sequence items must have a 'read' method (file-like
        object) that is called to fetch the bytes in memory.
        Otherwise the input is expected to be the sequence strings or
        bytes items are expected to be analyzed directly.
    encoding : string, 'utf-8' by default.
        If bytes or files are given to analyze, this encoding is used to
        decode.
    decode_error : {'strict', 'ignore', 'replace'}
        Instruction on what to do if a byte sequence is given to analyze that
        contains characters not of the given `encoding`. By default, it is
        'strict', meaning that a UnicodeDecodeError will be raised. Other
        values are 'ignore' and 'replace'.
    strip_accents : {'ascii', 'unicode', None}
        Remove accents during the preprocessing step.
        'ascii' is a fast method that only works on characters that have
        an direct ASCII mapping.
        'unicode' is a slightly slower method that works on any characters.
        None (default) does nothing.
    analyzer : string, {'word', 'char'} or callable
        Whether the feature should be made of word or character n-grams.
        If a callable is passed it is used to extract the sequence of features
        out of the raw, unprocessed input.
    preprocessor : callable or None (default)
        Override the preprocessing (string transformation) stage while
        preserving the tokenizing and n-grams generation steps.
    tokenizer : callable or None (default)
        Override the string tokenization step while preserving the
        preprocessing and n-grams generation steps.
        Only applies if ``analyzer == 'word'``.
    ngram_range : tuple (min_n, max_n)
        The lower and upper boundary of the range of n-values for different
        n-grams to be extracted. All values of n such that min_n <= n <= max_n
        will be used.
    stop_words : string {'english'}, list, or None (default)
        If a string, it is passed to _check_stop_list and the appropriate stop
        list is returned. 'english' is currently the only supported string
        value.
        If a list, that list is assumed to contain stop words, all of which
        will be removed from the resulting tokens.
        Only applies if ``analyzer == 'word'``.
        If None, no stop words will be used. max_df can be set to a value
        in the range [0.7, 1.0) to automatically detect and filter stop
        words based on intra corpus document frequency of terms.
    lowercase : boolean, default True
        Convert all characters to lowercase before tokenizing.
    token_pattern : string
        Regular expression denoting what constitutes a "token", only used
        if ``analyzer == 'word'``. The default regexp selects tokens of 2
        or more alphanumeric characters (punctuation is completely ignored
        and always treated as a token separator).
    max_df : float in range [0.0, 1.0] or int, default=1.0
        When building the vocabulary ignore terms that have a document
        frequency strictly higher than the given threshold (corpus-specific
        stop words).
        If float, the parameter represents a proportion of documents, integer
        absolute counts.
        This parameter is ignored if vocabulary is not None.
    min_df : float in range [0.0, 1.0] or int, default=1
        When building the vocabulary ignore terms that have a document
        frequency strictly lower than the given threshold. This value is also
        called cut-off in the literature.
        If float, the parameter represents a proportion of documents, integer
        absolute counts.
        This parameter is ignored if vocabulary is not None.
    max_features : int or None, default=None
        If not None, build a vocabulary that only consider the top
        max_features ordered by term frequency across the corpus.
        This parameter is ignored if vocabulary is not None.
    vocabulary : Mapping or iterable, optional
        Either a Mapping (e.g., a dict) where keys are terms and values are
        indices in the feature matrix, or an iterable over terms. If not
        given, a vocabulary is determined from the input documents.
    binary : boolean, default=False
        If True, all non-zero term counts are set to 1. This does not mean
        outputs will have only 0/1 values, only that the tf term in tf-idf
        is binary. (Set idf and normalization to False to get 0/1 outputs.)
    dtype : type, optional
        Type of the matrix returned by fit_transform() or transform().
    norm : 'l1', 'l2' or None, optional
        Norm used to normalize term vectors. None for no normalization.
    use_idf : boolean, default=True
        Enable inverse-document-frequency reweighting.
    smooth_idf : boolean, default=True
        Smooth idf weights by adding one to document frequencies, as if an
        extra document was seen containing every term in the collection
        exactly once. Prevents zero divisions.
    sublinear_tf : boolean, default=False
        Apply sublinear tf scaling, i.e. replace tf with 1 + log(tf).

    Attributes
    ----------
    vocabulary_ : dict
        A mapping of terms to feature indices.
    idf_ : array, shape = [n_features], or None
        The learned idf vector (global term weights)
        when ``use_idf`` is set to True, None otherwise.
    stop_words_ : set
        Terms that were ignored because they either:
          - occurred in too many documents (`max_df`)
          - occurred in too few documents (`min_df`)
          - were cut off by feature selection (`max_features`).
        This is only available if no vocabulary was given.
    """

    def __init__(self, input='content', encoding='utf-8',
                 decode_error='strict', strip_accents=None, lowercase=True,
                 preprocessor=None, tokenizer=None, analyzer='word',
                 stop_words=None, token_pattern=r"(?u)\b\w\w+\b",
                 ngram_range=(1, 1), max_df=1.0, min_df=1,
                 max_features=None, vocabulary=None, binary=False,
                 dtype=np.int64, norm='l2', use_idf=True, smooth_idf=True,
                 sublinear_tf=False):
        super(TfidfVectorizer, self).__init__(input, encoding, decode_error,
                                              strip_accents, lowercase,
                                              preprocessor, tokenizer, analyzer,
                                              stop_words, token_pattern,
                                              ngram_range, max_df, min_df,
                                              max_features, vocabulary, binary,
                                              dtype, norm, use_idf, smooth_idf,
                                              sublinear_tf)

    def fit_transform(self, raw_documents, y=None):
        """Learn vocabulary and idf, return term-document matrix.
        This is equivalent to fit followed by transform, but more efficiently
        implemented.
        Parameters
        ----------
        raw_documents : iterable
            an iterable which yields either str, unicode or file objects
        Returns
        -------
        X : sparse matrix, [n_samples, n_features]
            Tf-idf-weighted document-term matrix.
        """
        documents = super(TfidfVectorizer, self).fit_transform(
            raw_documents=raw_documents, y=y)
        count = CountVectorizer(encoding=self.encoding,
                                decode_error=self.decode_error,
                                strip_accents=self.strip_accents,
                                lowercase=self.lowercase,
                                preprocessor=self.preprocessor,
                                tokenizer=self.tokenizer,
                                stop_words=self.stop_words,
                                token_pattern=self.token_pattern,
                                ngram_range=self.ngram_range,
                                analyzer=self.analyzer,
                                max_df=self.max_df,
                                min_df=self.min_df,
                                max_features=self.max_features,
                                vocabulary=self.vocabulary_,
                                binary=self.binary,
                                dtype=self.dtype)
        count.fit_transform(raw_documents=raw_documents, y=y)
        self.period_ = count.period_
        self.df_ = count.df_
        self.n = count.n
        return documents



import re

from underthesea.transformer.tagged_feature import functions


class TaggedTransformer:
    def __init__(self, templates=None):
        self.templates = [self._extract_template(template) for template in templates]

    def _extract_template(self, template):
        token_syntax = template
        matched = re.match(
            "T\[(?P<index1>\-?\d+)(\,(?P<index2>\-?\d+))?\](\[(?P<column>.*)\])?(\.(?P<function>.*))?",
            template,
        )
        column = matched.group("column")
        column = int(column) if column else 0
        index1 = int(matched.group("index1"))
        index2 = matched.group("index2")
        index2 = int(index2) if index2 else None
        func = matched.group("function")
        return index1, index2, column, func, token_syntax

    def word2features(self, s):
        features = []
        for i, token in enumerate(s):
            tmp = []
            for template in self.templates:
                index1, index2, column, func, token_syntax = template
                prefix = "%s=" % token_syntax

                if i + index1 < 0:
                    result = "%sBOS" % prefix
                    tmp.append(result)
                    continue
                if i + index1 >= len(s):
                    result = "%sEOS" % prefix
                    tmp.append(result)
                    continue
                if index2 is not None:
                    if i + index2 >= len(s):
                        result = "%sEOS" % prefix
                        tmp.append(result)
                        continue
                    tokens = [s[j][column] for j in range(i + index1, i + index2 + 1)]
                    word = " ".join(tokens)
                else:
                    try:
                        word = s[i + index1][column]
                    except Exception:
                        pass
                if func is not None:
                    result = functions[func](word)
                else:
                    result = word
                result = "%s%s" % (prefix, result)
                tmp.append(result)
            features.append(tmp)
        return features

    def transform(self, sentences, contain_labels=False):
        X = [self.word2features(sentence) for sentence in sentences]
        if contain_labels:
            y = [self.sentence2labels(s) for s in sentences]
            return X, y
        return X

    def process(self, sentences, contain_labels=False):
        X = [self.word2features(sentence) for sentence in sentences]
        if contain_labels:
            y = [self.sentence2labels(s) for s in sentences]
            return X, y
        return X

    def sentence2labels(self, s):
        return [row[-1] for row in s]



import openai
import os


openai.api_key = os.environ.get('OPENAI_API_KEY')
if not openai.api_key:
    raise ValueError("Please set the OPENAI_API_KEY environment variable.")


def classify(X, domain=None):
    class_labels = "Thể thao, Pháp luật, Thế giới, Đời sống, Chính trị Xã hội, Vi tính, Khoa học, Văn hoá, Kinh doanh, Sức khỏe"
    user_prompt = f"""Classify the following text:
    {X}
    Suggested labels: {class_labels}

    Provide a single label as the output.
    """
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo",  # this may need to be changed based on available models
        messages=[
            {"role": "system", "content": "You are a helpful assistant."},
            {"role": "user", "content": user_prompt}
        ],
        temperature=0
    )
    output = response.choices[0].message['content'].strip()
    return output



import json
import warnings

# ignore warnings when using transformer
# see: https://github.com/scikit-learn/scikit-learn/issues/12327
from underthesea.corpus import Corpus
from underthesea.models.text_classifier import TextClassifier, TEXT_CLASSIFIER_ESTIMATOR
from sklearn.metrics import f1_score
from sklearn.preprocessing import LabelEncoder
from os.path import join
import joblib

warnings.simplefilter("ignore", category=PendingDeprecationWarning)


class ClassifierTrainer:

    def __init__(self, classifier: TextClassifier, corpus: Corpus):
        self.classifier = classifier
        self.corpus = corpus

    def train(self, model_folder: str, scoring=f1_score):
        score = {}
        multilabel = self.classifier.multilabel
        metadata = {"estimator": self.classifier.estimator.value, "multilabel": multilabel}
        train, dev, test = self._convert_corpus(self.corpus, multilabel=multilabel)
        X_train, y_train = train
        X_dev, y_dev = dev
        X_test, y_test = test
        # if self.classifier.estimator == TEXT_CLASSIFIER_ESTIMATOR.FAST_TEXT:
        #     hyper_params = self.classifier.params
        #     tmp_data_folder = tempfile.mkdtemp()
        #
        #     self.corpus.save(tmp_data_folder)
        #     train_file = Path(tmp_data_folder) / "train.txt"
        #     model = fastText.train_supervised(input=str(train_file), **hyper_params)
        #     y_dev_pred = [item[0].replace("__label__", "") for item in model.predict(X_dev)[0]]
        #     y_dev = [s.labels[0].value for s in self.corpus.dev]
        #     dev_score = scoring(y_dev, y_dev_pred)
        #
        #     y_test_pred = [item[0].replace("__label__", "") for item in model.predict(X_test)[0]]
        #     y_test = [s.labels[0].value for s in self.corpus.dev]
        #     test_score = scoring(y_test, y_test_pred)
        #
        #     shutil.rmtree(tmp_data_folder)
        #
        #     path_to_file = join(model_folder, "model.bin")
        #     model.save_model(path_to_file)
        #     print(f"Model is saved in {path_to_file}")
        #
        #     print("Dev score:", dev_score)
        #     print("Test score:", test_score)
        #     score["dev_score"] = dev_score
        #     score["test_score"] = test_score

        if self.classifier.estimator == TEXT_CLASSIFIER_ESTIMATOR.SVC:
            transformer = self.classifier.params['vectorizer']

            X_train = transformer.fit_transform(X_train)
            joblib.dump(transformer, join(model_folder, "x_transformer.joblib"))

            y_transformer = LabelEncoder()
            y_train = y_transformer.fit_transform(y_train)
            joblib.dump(y_transformer, join(model_folder, "y_transformer.joblib"))

            estimator = self.classifier.params['svc']
            estimator.fit(X_train, y_train)
            joblib.dump(estimator, join(model_folder, "estimator.joblib"))

            X_dev = transformer.transform(X_dev)
            y_dev = y_transformer.transform(y_dev)
            y_dev_pred = estimator.predict(X_dev)
            dev_score = scoring(y_dev, y_dev_pred)

            X_test = transformer.transform(X_test)
            y_test = y_transformer.transform(y_test)
            y_test_pred = estimator.predict(X_test)
            test_score = scoring(y_test, y_test_pred)
            score["dev_score"] = dev_score
            score["test_score"] = test_score
            print("Dev score:", dev_score)
            print("Test score:", test_score)

        if self.classifier.estimator == TEXT_CLASSIFIER_ESTIMATOR.PIPELINE:
            pipeline = self.classifier.pipeline
            if self.classifier.multilabel:
                y_train = self.classifier.y_encoder.fit_transform(y_train)
                joblib.dump(self.classifier.y_encoder, join(model_folder, "y_encoder.joblib"))
            pipeline.fit(X_train, y_train)
            joblib.dump(pipeline, join(model_folder, "pipeline.joblib"))

            y_dev_pred = pipeline.predict(X_dev)
            if self.classifier.multilabel:
                dev_score = scoring(self.classifier.y_encoder.transform(y_dev), y_dev_pred)
            else:
                dev_score = scoring(y_dev, y_dev_pred)

            y_test_pred = pipeline.predict(X_test)
            if self.classifier.multilabel:
                test_score = scoring(self.classifier.y_encoder.transform(y_test), y_test_pred)
            else:
                test_score = scoring(y_test, y_test_pred)
            score["dev_score"] = dev_score
            score["test_score"] = test_score
            print("Dev score:", dev_score)
            print("Test score:", test_score)

        with open(join(model_folder, "metadata.json"), "w") as f:
            content = json.dumps(metadata, ensure_ascii=False)
            f.write(content)
        return score

    def _convert_corpus(self, corpus: Corpus, multilabel=False):
        X_train = [s.text for s in corpus.train]
        X_dev = [s.text for s in corpus.dev]
        X_test = [s.text for s in corpus.test]
        if multilabel:
            y_train = [[label.value for label in s.labels] for s in corpus.train]
            y_dev = [[label.value for label in s.labels] for s in corpus.dev]
            y_test = [[label.value for label in s.labels] for s in corpus.test]
        else:
            y_train = [s.labels[0].value for s in corpus.train]
            y_dev = [s.labels[0].value for s in corpus.dev]
            y_test = [s.labels[0].value for s in corpus.test]
        return (X_train, y_train), (X_dev, y_dev), (X_test, y_test)



import os
from os import listdir
from os.path import join

from underthesea.file_utils import UNDERTHESEA_FOLDER
from underthesea.utils import logger
from underthesea.utils.col_script import UDDataset, UDSentence

WIKI_FOLDER = join(UNDERTHESEA_FOLDER, "data", "viwiki-20210720")
CLEANED_FOLDER = join(WIKI_FOLDER, "cleaned", "AA")
UD_FOLDER = join(WIKI_FOLDER, "ud", "AA")
os.makedirs(UD_FOLDER, exist_ok=True)


def check_line(line):
    if len(line) < 30:
        return False
    if line.startswith("<"):
        return False
    return True


def make_ud_file(file):
    logger.info(msg=file)
    sentences = []
    i = 0
    for line in open(join(CLEANED_FOLDER, file)):
        s = UDSentence.load_from_raw_text(line)
        sentences.append(s)
        i += 1
        if i % 200 == 0:
            logger.info(f"{file}:{i}")
    ud_dataset = UDDataset(sentences)
    ud_dataset.write(join(UD_FOLDER, file))
    logger.info(f"Finish {file}")


if __name__ == '__main__':
    files = sorted(listdir(CLEANED_FOLDER))

    # with Pool(5) as p:
    #     p.map(make_ud_file, files)

    # file = files[0]
    # make_ud_file(file)

    for file in files:
        make_ud_file(file)



from transformers import pipeline, AutoModelForTokenClassification, AutoTokenizer

model_repo = "undertheseanlp/vietnamese-ner-v1.4.0a2"
model_fine_tuned = AutoModelForTokenClassification.from_pretrained(model_repo)
tokenizer = AutoTokenizer.from_pretrained(model_repo)
nlp = pipeline("ner", model=model_fine_tuned, tokenizer=tokenizer)

if __name__ == '__main__':
    example = "Chưa tiết lộ lịch trình tới Việt Nam của Tổng thống Mỹ Donald Trump"
    ner_results = nlp(example)
    for entity in ner_results:
        print(entity)



import sys

if sys.version_info >= (3, 0):
    unicode = str


class UnicodeTransformer:
    def __init__(self):
        pass

    def transform(self, text):
        """
        :param unicode|str text: input text
        :type text: unicode|str

        :return: transformed text
        :rtype: unicode
        """
        if type(text) is not unicode:
            text = unicode(text, 'utf-8')
        return text



import re
import unicodedata
from argparse import ArgumentParser
from pathlib import Path

import soundfile as sf

from .hifigan.mel2wave import mel2wave
from .nat.config import FLAGS
from .nat.text2mel import text2mel


def nat_normalize_text(text):
    text = unicodedata.normalize("NFKC", text)
    text = text.lower().strip()
    sil = FLAGS.special_phonemes[FLAGS.sil_index]
    text = re.sub(r"[\n.,:]+", f" {sil} ", text)
    text = text.replace('"', " ")
    text = re.sub(r"\s+", " ", text)
    text = re.sub(r"[.,:;?!]+", f" {sil} ", text)
    text = re.sub("[ ]+", " ", text)
    text = re.sub(f"( {sil}+)+ ", f" {sil} ", text)
    return text.strip()


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--text", type=str)
    parser.add_argument("--output", default="clip.wav", type=Path)
    parser.add_argument("--sample-rate", default=16000, type=int)
    parser.add_argument("--silence-duration", default=-1, type=float)
    parser.add_argument("--lexicon-file", default=None)
    args = parser.parse_args()

    text = nat_normalize_text(args.text)
    print("Normalized text input:", text)
    mel = text2mel(text, args.lexicon_file, args.silence_duration)
    wave = mel2wave(mel)
    print("writing output to file", args.output)
    sf.write(str(args.output), wave, samplerate=args.sample_rate)



from os import remove
from urllib.request import urlretrieve
from os.path import dirname, join, exists
from zipfile import ZipFile

components = [
    {
        "name": "classification.vntc.model",
        "url": "https://www.dropbox.com/sh/y3wtsw4v8x0z5ll/AADLpHIdUW9LD7WL5_CzslRKa?dl=1",
        "zip_destination": ["classification/vntc", "tc_svm_vntc_20190607.zip"],
        "model_destination": ["classification/vntc", "tc_svm_vntc_20190607"]
    }
]


def download_component(component_name):
    try:
        component = [component for component in components if
                     component["name"] == component_name][0]
        try:
            folder = dirname(dirname(__file__))
            zip_folder = join(folder, join(*component["zip_destination"]))
            model_folder = join(folder, join(*component["model_destination"]))
            if exists(model_folder):
                print("Component '{}' is already existed.".format(component["name"]))
            else:
                print("Start download component '{}'".format(component["name"]))
                print(zip_folder)
                download_file(component["url"], zip_folder)
                ZipFile(zip_folder).extractall(model_folder)
                remove(zip_folder)
                print("Finish download component '{}'".format(component["name"]))
        except Exception as e:
            print(e)
            print("Cannot download component '{}'".format(component["name"]))
    except Exception:
        message = "Error: Component with name '{}' does not exist.".format(
            component_name)
        print(message)


def download_file(url, file_name):
    """ Cross platform file download helper function
    :param url: url of file
    :param file_name: destination file
    """
    urlretrieve(url, file_name)


if __name__ == '__main__':
    pass



import logging
import os
import sys
from os.path import dirname
from underthesea.corpus.data import Sentence
from underthesea.models.text_classifier import TextClassifier
from underthesea.model_fetcher import ModelFetcher, UTSModel

FORMAT = '%(message)s'
logging.basicConfig(format=FORMAT)
logger = logging.getLogger('underthesea')

sys.path.insert(0, dirname(dirname(__file__)))
model_path = ModelFetcher.get_model_path(UTSModel.tc_bank)
classifier = None

sys.path.insert(0, dirname(dirname(__file__)))
classifier = None


def sentiment(X):
    global classifier
    model_name = 'SA_BANK_V131'
    model_path = ModelFetcher.get_model_path(model_name)

    if not classifier:
        if not os.path.exists(model_path):
            ModelFetcher.download(model_name)
        classifier = TextClassifier.load(model_path)

    sentence = Sentence(X)
    classifier.predict(sentence)
    labels = sentence.labels
    if not labels:
        return None
    labels = [label.value for label in labels]
    return labels



from os.path import dirname, join
import pandas as pd

PROJECT_FOLDER = dirname(dirname(dirname(__file__)))
DATASETS_FOLDER = join(PROJECT_FOLDER, "datasets")
COL_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL")
DICTIONARY_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL", "dictionary")
DICTIONARY_DATA_FOLDER = join(DICTIONARY_FOLDER, "data")

entry_file = join(DICTIONARY_DATA_FOLDER, "entry.xlsx")


writer = pd.ExcelWriter(entry_file, engine='xlsxwriter')

workbook = writer.book
ws = workbook.add_worksheet('Sheet1')
ws.set_column(0, 0, 20)
ws.set_column(1, 1, 20)
ws.write(0, 0, 'abc')
ws.write(0, 0, 'HEADWORD')
ws.write('A3', 'SENSE #1')
ws.write('A4', 'word class')
ws.write('A5', 'examples')

writer.save()



# -*- coding: utf-8 -*-
from .regex_tokenize import tokenize
from underthesea.models.fast_crf_sequence_tagger import FastCRFSequenceTagger
from os.path import dirname, join

word_tokenize_model = None


def word_tokenize(sentence, format=None, use_token_normalize=True, fixed_words=[]):
    """
    Vietnamese word segmentation

    Args:
        sentence (str): raw sentence
        format (str, optional): format option.
            Defaults to None.
            use format=`text` for text format
        use_token_normalize (bool): True if use token_normalize
        fixed_words (list): list of fixed words

    Returns:
        :obj:`list` of :obj:`str`:
            word tokens

    Examples:

        >>> # -*- coding: utf-8 -*-
        >>> from underthesea import word_tokenize
        >>> sentence = "Bác sĩ bây giờ có thể thản nhiên báo tin bệnh nhân bị ung thư"

        >>> word_tokenize(sentence)
        ["Bác sĩ", "bây giờ", "có thể", "thản nhiên", "báo tin", "bệnh nhân", "bị", "ung thư"]

        >>> word_tokenize(sentence, format="text")
        "Bác_sĩ bây_giờ có_thể thản_nhiên báo_tin bệnh_nhân bị ung_thư"
    """
    global word_tokenize_model
    tokens = tokenize(sentence, use_token_normalize=use_token_normalize, fixed_words=fixed_words)
    features = [[token] for token in tokens]
    if word_tokenize_model is None:
        word_tokenize_model = FastCRFSequenceTagger()
        wd = dirname(__file__)
        word_tokenize_model.load(join(wd, "models", "ws_crf_vlsp2013_20230727"))
    tags = word_tokenize_model.predict(features)

    output = []
    num_words = 0
    for tag, token in zip(tags, tokens):
        if tag == "I-W" and num_words > 0:
            output[-1] = output[-1] + u" " + token
        else:
            output.append(token)
        num_words += 1
    if format == "text":
        output = u" ".join([item.replace(" ", "_") for item in output])
    return output



import io


def read(filename):
    with io.open(filename, "r", encoding="utf-8") as f:
        text = f.read()
    return text


def write(filename, text):
    with io.open(filename, "w", encoding="utf-8") as f:
        f.write(text)



import haiku as hk
import jax
import jax.numpy as jnp
from jax.numpy import ndarray

from .config import FLAGS, AcousticInput, DurationInput


class TokenEncoder(hk.Module):
    """Encode phonemes/text to vector"""

    def __init__(self, vocab_size, lstm_dim, dropout_rate, is_training=True):
        super().__init__()
        self.is_training = is_training
        self.embed = hk.Embed(vocab_size, lstm_dim)
        self.conv1 = hk.Conv1D(lstm_dim, 3, padding="SAME")
        self.conv2 = hk.Conv1D(lstm_dim, 3, padding="SAME")
        self.conv3 = hk.Conv1D(lstm_dim, 3, padding="SAME")
        self.bn1 = hk.BatchNorm(True, True, 0.9)
        self.bn2 = hk.BatchNorm(True, True, 0.9)
        self.bn3 = hk.BatchNorm(True, True, 0.9)
        self.lstm_fwd = hk.LSTM(lstm_dim)
        self.lstm_bwd = hk.ResetCore(hk.LSTM(lstm_dim))
        self.dropout_rate = dropout_rate

    def __call__(self, x, lengths):
        x = self.embed(x)
        x = jax.nn.relu(self.bn1(self.conv1(x), is_training=self.is_training))
        if self.is_training:
            x = hk.dropout(hk.next_rng_key(), self.dropout_rate, x)
        x = jax.nn.relu(self.bn2(self.conv2(x), is_training=self.is_training))
        if self.is_training:
            x = hk.dropout(hk.next_rng_key(), self.dropout_rate, x)
        x = jax.nn.relu(self.bn3(self.conv3(x), is_training=self.is_training))
        if self.is_training:
            x = hk.dropout(hk.next_rng_key(), self.dropout_rate, x)
        B, L, _ = x.shape
        mask = jnp.arange(0, L)[None, :] >= (lengths[:, None] - 1)
        h0c0_fwd = self.lstm_fwd.initial_state(B)
        new_hx_fwd, _ = hk.dynamic_unroll(self.lstm_fwd, x, h0c0_fwd, time_major=False)
        x_bwd, mask_bwd = jax.tree_map(lambda x: jnp.flip(x, axis=1), (x, mask))
        h0c0_bwd = self.lstm_bwd.initial_state(B)
        new_hx_bwd, _ = hk.dynamic_unroll(
            self.lstm_bwd, (x_bwd, mask_bwd), h0c0_bwd, time_major=False
        )
        x = jnp.concatenate((new_hx_fwd, jnp.flip(new_hx_bwd, axis=1)), axis=-1)
        return x


class DurationModel(hk.Module):
    """Duration model of phonemes."""

    def __init__(self, is_training=True):
        super().__init__()
        self.is_training = is_training
        self.encoder = TokenEncoder(
            FLAGS.vocab_size,
            FLAGS.duration_lstm_dim,
            FLAGS.duration_embed_dropout_rate,
            is_training,
        )
        self.projection = hk.Sequential(
            [hk.Linear(FLAGS.duration_lstm_dim), jax.nn.gelu, hk.Linear(1)]
        )

    def __call__(self, inputs: DurationInput):
        x = self.encoder(inputs.phonemes, inputs.lengths)
        x = jnp.squeeze(self.projection(x), axis=-1)
        x = jax.nn.softplus(x)
        return x


class AcousticModel(hk.Module):
    """Predict melspectrogram from aligned phonemes"""

    def __init__(self, is_training=True):
        super().__init__()
        self.is_training = is_training
        self.encoder = TokenEncoder(
            FLAGS.vocab_size, FLAGS.acoustic_encoder_dim, 0.5, is_training
        )
        self.decoder = hk.deep_rnn_with_skip_connections(
            [hk.LSTM(FLAGS.acoustic_decoder_dim), hk.LSTM(FLAGS.acoustic_decoder_dim)]
        )
        self.projection = hk.Linear(FLAGS.mel_dim)

        # prenet
        self.prenet_fc1 = hk.Linear(256, with_bias=False)
        self.prenet_fc2 = hk.Linear(256, with_bias=False)
        # posnet
        self.postnet_convs = [hk.Conv1D(FLAGS.postnet_dim, 5) for _ in range(4)]
        self.postnet_convs.append(hk.Conv1D(FLAGS.mel_dim, 5))
        self.postnet_bns = [hk.BatchNorm(True, True, 0.9) for _ in range(4)] + [None]

    def prenet(self, x, dropout=0.5):
        x = jax.nn.relu(self.prenet_fc1(x))
        x = hk.dropout(hk.next_rng_key(), dropout, x)
        x = jax.nn.relu(self.prenet_fc2(x))
        x = hk.dropout(hk.next_rng_key(), dropout, x)
        return x

    def upsample(self, x, durations, L):
        ruler = jnp.arange(0, L)[None, :]  # B, L
        end_pos = jnp.cumsum(durations, axis=1)
        mid_pos = end_pos - durations / 2  # B, T

        d2 = jnp.square((mid_pos[:, None, :] - ruler[:, :, None])) / 10.0
        w = jax.nn.softmax(-d2, axis=-1)
        hk.set_state("attn", w[0])
        x = jnp.einsum("BLT,BTD->BLD", w, x)
        return x

    def postnet(self, mel: ndarray) -> ndarray:
        x = mel
        for conv, bn in zip(self.postnet_convs, self.postnet_bns):
            x = conv(x)
            if bn is not None:
                x = bn(x, is_training=self.is_training)
                x = jnp.tanh(x)
            x = hk.dropout(hk.next_rng_key(), 0.5, x) if self.is_training else x
        return x

    def inference(self, tokens, durations, n_frames):
        B, L = tokens.shape
        lengths = jnp.array([L], dtype=jnp.int32)
        x = self.encoder(tokens, lengths)
        x = self.upsample(x, durations, n_frames)

        def loop_fn(inputs, state):
            cond = inputs
            prev_mel, hxcx = state
            prev_mel = self.prenet(prev_mel)
            x = jnp.concatenate((cond, prev_mel), axis=-1)
            x, new_hxcx = self.decoder(x, hxcx)
            x = self.projection(x)
            return x, (x, new_hxcx)

        state = (
            jnp.zeros((B, FLAGS.mel_dim), dtype=jnp.float32),
            self.decoder.initial_state(B),
        )
        x, _ = hk.dynamic_unroll(loop_fn, x, state, time_major=False)
        residual = self.postnet(x)
        return x + residual

    def __call__(self, inputs: AcousticInput):
        x = self.encoder(inputs.phonemes, inputs.lengths)
        x = self.upsample(x, inputs.durations, inputs.mels.shape[1])
        mels = self.prenet(inputs.mels)
        x = jnp.concatenate((x, mels), axis=-1)
        B, L, _ = x.shape
        hx = self.decoder.initial_state(B)

        def zoneout_decoder(inputs, prev_state):
            x, mask = inputs
            x, state = self.decoder(x, prev_state)
            state = jax.tree_multimap(
                lambda m, s1, s2: s1 * m + s2 * (1 - m), mask, prev_state, state
            )
            return x, state

        mask = jax.tree_map(
            lambda x: jax.random.bernoulli(hk.next_rng_key(), 0.1, (B, L, x.shape[-1])),
            hx,
        )
        x, _ = hk.dynamic_unroll(zoneout_decoder, (x, mask), hx, time_major=False)
        x = self.projection(x)
        residual = self.postnet(x)
        return x, x + residual



import pickle
from os.path import join, dirname


class UVD:
    """
    Underthesea Vietnamese Dictionary
    """
    version = '0.0.1'
    date = '2022/03/05'

    def __init__(self):
        """ contains 13k words in this version
        """
        filepath = join(dirname(__file__), "dictionary.data")
        self.words = []
        with open(filepath, "rb") as f:
            self.words = pickle.load(f)


class DictionaryStat:
    @staticmethod
    def stat(dictionary):
        print('Words:', len(dictionary.words))



from .character_normalize import normalize_characters_in_text
from .text_normalizer import token_map


def token_normalize(token, use_character_normalize=True):
    """
    normalize each token
    """
    if len(token) > 6:
        return token
    # character normalize
    if use_character_normalize:
        token = normalize_characters_in_text(token)
    if token in token_map:
        return token_map[token]
    return token



import os
from math import log
from os.path import dirname, join
from collections import defaultdict
from underthesea.utils.col_script import UDDataset
from underthesea.utils.col_analyzer import UDAnalyzer

BOT_VERSION = "underthesea.v1.3.2"
PROJECT_FOLDER = dirname(dirname(dirname((os.getcwd()))))
DATASETS_FOLDER = join(PROJECT_FOLDER, "datasets")
COL_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL")
STOPWORDS_FOLDER = join(PROJECT_FOLDER, "underthesea", "datasets", "stopwords")


def invert_index(doc_word_counters):
    """Get word as key, all docs containing word as value"""
    index = defaultdict(list)
    for doc, counter in doc_word_counters.items():
        for word in counter.keys():
            index[word].append(doc)
    return index


def kl_div(p_x, p_c):
    return p_x * log(p_x / p_c, 2)


class KLDivergence:
    def __init__(self, word_counter, doc_word_counters):
        """
        F: Term (word) frequency
        L_x: Frequency sum of all docs containing word
        P_c: Distribution of word in the collection
        P_x: Distribution of word in sampled docs
        """
        self.words = word_counter.keys()
        self.F = [word_counter.get(word) for word in self.words]
        self.index = invert_index(doc_word_counters)
        self.L_x = [sum(doc_word_counters[doc].values()) for word in self.words for doc in self.index[word]]
        self.P_c = [float(F) / sum(word_counter.values()) for F in self.F]
        self.P_x = [float(self.F[i]) / self.L_x[i] for i in range(len(self.F))]
        self.values = [kl_div(self.P_x[i], self.P_c[i]) for i in range(len(self.P_x))]

    def write(self):
        words = [x for _, x in sorted(zip(self.values, self.words), reverse=True)][:70]  # Top 70
        content = "\n".join([s for s in words])
        target_file = join(STOPWORDS_FOLDER, "stopwords_dev.txt")
        with open(target_file, "w") as f:
            f.write(content)


def main():
    raw_file = join(COL_FOLDER, "corpus", "raw", "202108.txt")
    dataset = UDDataset.load_from_raw_file(raw_file)

    analyzer = UDAnalyzer()
    word_counter = analyzer.analyze_words(dataset)
    doc_word_counters = analyzer.get_doc_word_counters(dataset)

    KLDivergence(word_counter, doc_word_counters).write()


if __name__ == "__main__":
    main()



import warnings
from abc import abstractmethod
from pathlib import Path
from typing import Union
import torch.nn

from underthesea import file_utils
from underthesea.utils import device


class Model(torch.nn.Module):
    r"""
    Abstract base class for all downstream task models
    Every new type of model must implement these methods.

    Source: FlairNLP
    """

    @abstractmethod
    def evaluate(self):
        r"""Evaluates the model. Returns a Result object containing evaluation
        """
        pass

    def save(self, model_file: Union[str, Path]):
        """
        Saves the current model to the provided file.

        Args:
            model_file (Union[str, Path]): the model file

        """
        model_state = self._get_state_dict()

        torch.save(model_state, str(model_file), pickle_protocol=4)

    @staticmethod
    @abstractmethod
    def _fetch_model(model_name) -> str:
        return model_name

    @staticmethod
    @abstractmethod
    def _init_model_with_state_dict(state):
        """Initialize the model from a state dictionary. Implementing this enables the load() and load_checkpoint()
        functionality."""
        pass

    @classmethod
    def load(cls, model: Union[str, Path]):
        """
        Loads the model from the given file.
        :param model: the model file
        :return: the loaded text classifier model
        """
        model_file = cls._fetch_model(str(model))

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore")
            # load_big_file is a workaround by https://github.com/highway11git to load models on some Mac/Windows setups
            # see https://github.com/zalandoresearch/flair/issues/351
            f = file_utils.load_big_file(str(model_file))
        state = torch.load(f, map_location='cpu')

        model = cls._init_model_with_state_dict(state)

        model.eval()
        model.to(device)

        return model



##################################################################
# PREPROCESS ORIGINAL CORPUS: VLSP2020-DP
##################################################################

##################################################################
# CONTENT
##################################################################
# d    4.0K    Jan 16 23:12    DataTestDP2020
# d    4.0K    Jan 16 23:12    DataTestDP2020-CoNLLU
# d    4.0K    Jan 16 23:12    DataTestGoldDP2020
# -    2.9M    Nov 18 09:44    DP-Package2.18.11.2020.txt
# -    665K    Sep 18 09:22    HTB_1570.txt
# -     337    Nov  8 23:09    MotaDulieuDot1.txt
# -     26K    Nov  8 23:07    SA-Hotel-50.txt
# -     20K    Nov  8 23:07    SA-Restaurant-50.txt
# -    1.6M    May 12  2020    VTB_2996.txt
# -    375K    Nov  8 23:07    VTB_400.txt
import os
import shutil
from os.path import join

from underthesea.file_utils import DATASETS_FOLDER

ORIGINAL_ZIP_FILE = join(DATASETS_FOLDER, 'VLSP2020-DP.zip')
ORIGINAL_FOLDER = join(DATASETS_FOLDER, 'VLSP2020-DP-ORIGINAL')
DEST_FOLDER = join(DATASETS_FOLDER, 'VLSP2020-DP')
shutil.rmtree(DEST_FOLDER)
os.makedirs(DEST_FOLDER)


# shutil.rmtree(ORIGINAL_FOLDER)
# with ZipFile(ORIGINAL_ZIP_FILE) as file:
#     file.extractall(ORIGINAL_FOLDER)

class CONLLReader:
    @staticmethod
    def load(file):
        sentences = open(file).read().split('\n\n')
        sentences = [s.strip() for s in sentences if len(s) > 10]
        return sentences

    @staticmethod
    def save(sentences, file):
        n = len(sentences)
        content = '\n\n'.join(sentences) + '\n'
        print(f'Save {n} sentences in {file}')
        with open(file, 'w') as f:
            f.write(content)


files = ['DP-Package2.18.11.2020.txt', 'VTB_2996.txt', 'VTB_400.txt', 'SA-Restaurant-50.txt', 'SA-Hotel-50.txt',
         'HTB_1570.txt']
files = [join(DATASETS_FOLDER, ORIGINAL_FOLDER, f) for f in files]
sentences = []
for file in files:
    sentences += CONLLReader.load(file)
CONLLReader.save(sentences, join(DEST_FOLDER, 'train.txt'))

file = join(ORIGINAL_FOLDER, 'DataTestGoldDP2020', 'total-gold.txt')
sentences = CONLLReader.load(file)
CONLLReader.save(sentences, join(DEST_FOLDER, 'test.txt'))



# flake8: noqa

from functools import partial
from typing import Optional

import jax
import jax.numpy as jnp
import librosa
from einops import rearrange
from jax.numpy import ndarray


def rolling_window(a: ndarray, window: int, hop_length: int):
    """return a stack of overlap subsequence of an array.
    ``return jnp.stack( [a[0:10], a[5:15], a[10:20],...], axis=0)``
    Source: https://github.com/google/jax/issues/3171
    Args:
      a (ndarray): input array of shape `[L, ...]`
      window (int): length of each subarray (window).
      hop_length (int): distance between neighbouring windows.
    """

    idx = (
        jnp.arange(window)[:, None] +
        jnp.arange((len(a) - window) // hop_length + 1)[None, :] * hop_length
    )
    return a[idx]


@partial(jax.jit, static_argnums=[1, 2, 3, 4, 5, 6])
def stft(
    y: ndarray,
    n_fft: int = 2048,
    hop_length: Optional[int] = None,
    win_length: Optional[int] = None,
    window: str = "hann",
    center: bool = True,
    pad_mode: str = "reflect",
):
    """A jax reimplementation of ``librosa.stft`` function."""

    if win_length is None:
        win_length = n_fft

    if hop_length is None:
        hop_length = win_length // 4

    if window == "hann":
        fft_window = jnp.hanning(win_length + 1)[:-1]
    else:
        raise RuntimeError(f"{window} window function is not supported!")

    pad_len = (n_fft - win_length) // 2
    fft_window = jnp.pad(fft_window, (pad_len, pad_len), mode="constant")
    fft_window = fft_window[:, None]
    if center:
        y = jnp.pad(y, int(n_fft // 2), mode=pad_mode)

    # jax does not support ``np.lib.stride_tricks.as_strided`` function
    # see https://github.com/google/jax/issues/3171 for comments.
    y_frames = rolling_window(y, n_fft, hop_length) * fft_window
    stft_matrix = jnp.fft.fft(y_frames, axis=0)
    d = int(1 + n_fft // 2)
    return stft_matrix[:d]


@partial(jax.jit, static_argnums=[1, 2, 3, 4, 5, 6])
def batched_stft(
    y: ndarray,
    n_fft: int,
    hop_length: int,
    win_length: int,
    window: str,
    center: bool = True,
    pad_mode: str = "reflect",
):
    """Batched version of ``stft`` function.
    TN => FTN
    """

    assert len(y.shape) >= 2
    if window == "hann":
        fft_window = jnp.hanning(win_length + 1)[:-1]
    else:
        raise RuntimeError(f"{window} window function is not supported!")
    pad_len = (n_fft - win_length) // 2
    if pad_len > 0:
        fft_window = jnp.pad(fft_window, (pad_len, pad_len), mode="constant")
        win_length = n_fft
    else:
        fft_window = fft_window
    if center:
        pad_width = ((n_fft // 2, n_fft // 2),) + ((0, 0),) * (len(y.shape) - 1)
        y = jnp.pad(y, pad_width, mode=pad_mode)

    # jax does not support ``np.lib.stride_tricks.as_strided`` function
    # see https://github.com/google/jax/issues/3171 for comments.
    y_frames = rolling_window(y, n_fft, hop_length)
    fft_window = jnp.reshape(fft_window, (-1,) + (1,) * (len(y.shape)))
    y_frames = y_frames * fft_window
    stft_matrix = jnp.fft.fft(y_frames, axis=0)
    d = int(1 + n_fft // 2)
    return stft_matrix[:d]


class MelFilter:
    """Convert waveform to mel spectrogram."""

    def __init__(self, sample_rate: int, n_fft: int, n_mels: int, fmin=0.0, fmax=8000):
        self.melfb = jax.device_put(
            librosa.filters.mel(
                sr=sample_rate, n_fft=n_fft, n_mels=n_mels, fmin=fmin, fmax=fmax
            )
        )
        self.n_fft = n_fft

    def __call__(self, y: ndarray) -> ndarray:
        hop_length = self.n_fft // 4
        window_length = self.n_fft
        assert len(y.shape) == 2
        y = rearrange(y, "n s -> s n")
        p = (self.n_fft - hop_length) // 2
        y = jnp.pad(y, ((p, p), (0, 0)), mode="reflect")
        spec = batched_stft(
            y, self.n_fft, hop_length, window_length, "hann", False, "reflect"
        )
        mag = jnp.sqrt(jnp.square(spec.real) + jnp.square(spec.imag) + 1e-9)
        mel = jnp.einsum("ms,sfn->nfm", self.melfb, mag)
        cond = jnp.log(jnp.clip(mel, a_min=1e-5, a_max=None))
        return cond



import os
from os.path import join

from underthesea.file_utils import UNDERTHESEA_FOLDER
from underthesea.utils.col_script import UDDataset


def convert_to_sketchengine(ud_dataset, filepath):
    sentences = []
    for s in ud_dataset[:1000]:
        rows_ = []
        for row in s.rows:
            word, tag, _, _ = row
            lemma = word.lower()
            row_content = "\t".join([word, tag, lemma])
            rows_.append(row_content)
        s_content = "<s>\n" + "\n".join(rows_) + "\n</s>"
        sentences.append(s_content)
    content = "\n".join(sentences)
    with open(filepath, 'w') as f:
        f.write(content)


if __name__ == '__main__':
    file = "wiki_00"
    ud_file = join(UNDERTHESEA_FOLDER, "data", "viwiki-20210720", "ud", "AA", file)
    ud_dataset = UDDataset.load(ud_file)

    WIKI_FOLDER = join(UNDERTHESEA_FOLDER, "data", "viwiki-20210720")
    # SE: sketch engine
    SE_FOLDER = join(WIKI_FOLDER, "se", "AA")
    os.makedirs(SE_FOLDER, exist_ok=True)
    se_file = join(SE_FOLDER, file + ".vert")

    convert_to_sketchengine(ud_dataset, se_file)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from underthesea.util import download_component


def download_default_components():
    # download default components
    default_components = ["classification.vntc.model"]
    [download_component(component) for component in default_components]


if __name__ == '__main__':
    download_default_components()



import io
from underthesea.corpus import Document, UnicodeTransformer, Corpus
from os import listdir, mkdir
from os.path import join

from underthesea.util.file_io import write


class WSCorpus(Corpus):
    def __init__(self):
        pass

    def map_token(self, token):
        if token[1] == "B_W":
            return " " + token[0]
        if token[1] == "I_W":
            return "_" + token[0]
        else:
            return " " + token[0]

    def _column_to_text(self, sentence):
        s = sentence
        tokens = s.split("\n")
        tokens = [t.split("\t") for t in tokens]
        tokens = [self.map_token(t) for t in tokens]
        s = "".join(tokens)
        s = s.strip()
        return s

    def load(self, folder, format):
        """

        :param str folder: path to directory
        :type folder: string
        :param str format: either TEXT or COLUMN
        :type format: str
        """
        ids = listdir(folder)
        files = [join(folder, file) for file in ids]
        contents = []
        for f in files:
            with io.open(f, "r", encoding="utf-8") as f:
                content = f.read().strip()
                contents.append(content)
        documents = []

        for id, content in zip(ids, contents):
            document = Document(id)
            unicode_transformer = UnicodeTransformer()
            content = unicode_transformer.transform(content)
            sentences = content.split("\n\n")
            sentences = [self._column_to_text(s) for s in sentences]
            document.set_sentences(sentences)
            documents.append(document)

        self.documents = documents

    def save(self, folder, format):
        """save wscorpus to files

        :param str folder: path to directory
        :type folder: string
        :param str format: either TEXT or COLUMN
        :type format: str
        """
        try:
            mkdir(folder)
        except Exception:
            pass
        for document in self.documents:
            f = join(folder, document.id)
            content = u"\n".join(document.sentences)
            write(f, content)



"""
Utilities for working with the local data set cache. Copied from flair
"""
import mmap
from os.path import join
from pathlib import Path
import os
import shutil
import tempfile
import re
from urllib.parse import urlparse
from tqdm import tqdm as _tqdm
import requests
from underthesea.utils import logger

UNDERTHESEA_FOLDER = os.path.expanduser(os.path.join('~', '.underthesea'))
DATASETS_FOLDER = join(UNDERTHESEA_FOLDER, "datasets")
MODELS_FOLDER = join(UNDERTHESEA_FOLDER, "models")


def cached_path(url_or_filename: str, cache_dir: Path) -> Path:
    """
    Given something that might be a URL (or might be a local path),
    determine which. If it's a URL, download the file and cache it, and
    return the path to the cached file. If it's already a local path,
    make sure the file exists and then return the path.
    """
    dataset_cache = Path(UNDERTHESEA_FOLDER) / cache_dir

    parsed = urlparse(url_or_filename)

    if parsed.scheme in ('http', 'https'):
        # URL, so get it from the cache (downloading if necessary)
        return get_from_cache(url_or_filename, dataset_cache)
    elif parsed.scheme == '' and Path(url_or_filename).exists():
        # File, and it exists.
        return Path(url_or_filename)
    elif parsed.scheme == '':
        # File, but it doesn't exist.
        raise FileNotFoundError("file {} not found".format(url_or_filename))
    else:
        # Something unknown
        raise ValueError("unable to parse {} as a URL or as a local path".format(url_or_filename))


# TODO(joelgrus): do we want to do checksums or anything like that?
def get_from_cache(url: str, cache_dir: Path = None) -> Path:
    """
    Given a URL, look for the corresponding dataset in the local cache.
    If it's not there, download it. Then return the path to the cached file.
    """
    cache_dir.mkdir(parents=True, exist_ok=True)

    filename = re.sub(r'.+/', '', url)
    # get cache path to put the file
    cache_path = cache_dir / filename
    if cache_path.exists():
        return cache_path

    # make HEAD request to check ETag
    response = requests.head(url)

    # (anhv: 27/12/2020) github release assets return 302
    if response.status_code not in [200, 302]:
        if "www.dropbox.com" in url:
            # dropbox return code 301, so we ignore this error
            pass
        else:
            raise IOError("HEAD request failed for url {}".format(url))

    # add ETag to filename if it exists
    # etag = response.headers.get("ETag")

    if not cache_path.exists():
        # Download to temporary file, then copy to cache dir once finished.
        # Otherwise you get corrupt cache entries if the download gets interrupted.
        fd, temp_filename = tempfile.mkstemp()
        logger.info("%s not found in cache, downloading to %s", url, temp_filename)

        # GET file object
        req = requests.get(url, stream=True)
        content_length = req.headers.get('Content-Length')
        total = int(content_length) if content_length is not None else None
        progress = Tqdm.tqdm(unit="B", total=total)
        with open(temp_filename, 'wb') as temp_file:
            for chunk in req.iter_content(chunk_size=1024):
                if chunk:  # filter out keep-alive new chunks
                    progress.update(len(chunk))
                    temp_file.write(chunk)

        progress.close()

        logger.info("copying %s to cache at %s", temp_filename, cache_path)
        shutil.copyfile(temp_filename, str(cache_path))
        logger.info("removing temp file %s", temp_filename)
        os.close(fd)
        os.remove(temp_filename)

    return cache_path


class Tqdm:
    # These defaults are the same as the argument defaults in tqdm.
    default_mininterval = 0.1

    @staticmethod
    def set_default_mininterval(value: float) -> None:
        Tqdm.default_mininterval = value

    @staticmethod
    def set_slower_interval(use_slower_interval: bool) -> None:
        """
        If ``use_slower_interval`` is ``True``, we will dramatically slow down ``tqdm's`` default
        output rate.  ``tqdm's`` default output rate is great for interactively watching progress,
        but it is not great for log files.  You might want to set this if you are primarily going
        to be looking at output through log files, not the terminal.
        """
        if use_slower_interval:
            Tqdm.default_mininterval = 10.0
        else:
            Tqdm.default_mininterval = 0.1

    @staticmethod
    def tqdm(*args, **kwargs):
        new_kwargs = {
            'mininterval': Tqdm.default_mininterval,
            **kwargs
        }

        return _tqdm(*args, **new_kwargs)


def load_big_file(f: str) -> mmap.mmap:
    r"""
    Workaround for loading a big pickle file. Files over 2GB cause pickle errors on certin Mac and Windows distributions.
    Source: flairNLP
    """
    logger.info(f"loading file {f}")
    with open(f, "rb") as f_in:
        # mmap seems to be much more memory efficient
        bf = mmap.mmap(f_in.fileno(), 0, access=mmap.ACCESS_READ)
        f_in.close()
    return bf



import pickle
from pathlib import Path

from tabulate import tabulate


def load_latest_ckpt(ckpt_dir: Path):
    ckpt = ckpt_dir / "duration_latest_ckpt.pickle"
    if not ckpt.exists():
        return None
    print("Loading latest checkpoint from file", ckpt)
    with open(ckpt, "rb") as f:
        dic = pickle.load(f)
    return dic["step"], dic["params"], dic["aux"], dic["rng"], dic["optim_state"]


def save_ckpt(step, params, aux, rng, optim_state, ckpt_dir: Path):
    dic = {
        "step": step,
        "params": params,
        "aux": aux,
        "rng": rng,
        "optim_state": optim_state,
    }
    with open(ckpt_dir / "duration_latest_ckpt.pickle", "wb") as f:
        pickle.dump(dic, f)


def print_flags(flags):
    values = [(k, v) for k, v in flags.items() if not k.startswith("_")]
    print(tabulate(values))



from os.path import join, dirname
from underthesea.util.file_io import read


class DictionaryLoader:
    def __init__(self, filepath):
        """load words from Ho Ngoc Duc's dictionary

        :param str filepath: filename of dictionary data
        :type filepath: str
        """
        data_folder = join(dirname(dirname(__file__)), "data")
        data_file = join(data_folder, filepath)
        self.data_file = data_file
        self.words_data = None

    @property
    def words(self):
        if not self.words_data:
            content = read(self.data_file).strip()
            words = content.split("\n")
            self.words_data = words
        return self.words_data



from os import listdir
from os.path import dirname, join
import pandas as pd
import yaml
from underthesea.file_utils import UNDERTHESEA_FOLDER
from underthesea.utils.col_external_dictionary import Cache, VLSPDictionary
from underthesea.utils.col_script import UDDataset

PROJECT_FOLDER = dirname(dirname(dirname(__file__)))
DATASETS_FOLDER = join(PROJECT_FOLDER, "datasets")
COL_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL")
DICTIONARY_FOLDER = join(DATASETS_FOLDER, "UD_Vietnamese-COL", "dictionary")
DICTIONARY_FILE = join(DICTIONARY_FOLDER, "202108.yaml")


class Word:
    pass


class Dictionary:
    def __init__(self, data):
        self.data = data

    def describe(self):
        print("Dictionary Describe")
        print(f"- Words: {len(self.data)}")

    def import_words(self, data_file):
        df = pd.read_excel(data_file, engine='openpyxl', nrows=5000)
        df = df[df['verify'] == 'x']
        for key, row in df.iterrows():
            token = row[0]
            pos = row[1]
            if token in self.data:
                senses = self.data[token]
                exist_pos = [s['tag'] for s in senses]
                if pos not in exist_pos:
                    self.data[token].append({'tag': pos})
            else:
                self.data[token] = [{'tag': pos}]

    def to_df(self):
        output = []
        for word in self.data:
            for sense in self.data[word]:
                item = {'token': word, 'pos': sense['tag']}
                output.append(item)
        df = pd.DataFrame(output)
        return df

    def save(self, dictionary_file):
        content = yaml.dump(self.data, allow_unicode=True, sort_keys=True)
        with open(dictionary_file, 'w') as f:
            f.write(content)

    @staticmethod
    def load(dictionary_file, cache_file=None):
        with open(DICTIONARY_FILE) as f:
            content = f.read()
        data = yaml.safe_load(content)
        return Dictionary(data)


if __name__ == '__main__':
    dictionary = Dictionary.load(DICTIONARY_FILE)
    dictionary.describe()
    current_df = dictionary.to_df()

    # ud_file = join(COL_FOLDER, "corpus", "ud", "202108.txt")
    ud_folder = join(UNDERTHESEA_FOLDER, "data", "viwiki-20210720", "ud", "AA")
    sentences = []
    for file in listdir(ud_folder)[:2]:
        ud_file = join(UNDERTHESEA_FOLDER, "data", "viwiki-20210720", "ud", "AA", file)
        ud_dataset = UDDataset.load(ud_file)
        sentences += ud_dataset.sentences
    ud_dataset = UDDataset(sentences)
    rows = [s.rows for s in ud_dataset]
    rows = [[row[0].lower(), row[1]] for sublist in rows for row in sublist]
    ud_df = pd.DataFrame(rows, columns=['token', 'pos'])
    print("List pos:", sorted(set(ud_df["pos"])))
    pos = [
        ["V", "verb"],
        ["N", "noun"],
        ["A", "adjective"],
        ["P", "pronoun"],
        ["E", "preposition"]
    ]
    vlsp_cache_file = join(DICTIONARY_FOLDER, "data", "vlsp_cache.bin")
    vlsp_cache = Cache.load(vlsp_cache_file)

    for pos_label, pos_full_name in pos:
        ud_df_sub = ud_df[ud_df["pos"].isin([pos_label])]
        ud_df_sub = ud_df_sub.groupby(["token", "pos"]).size().reset_index(name='count').sort_values("count",
                                                                                                     ascending=False)

        df = pd.merge(ud_df_sub, current_df, on=['token', 'pos'], how='outer', indicator=True)
        df = df[df['_merge'] == 'left_only']
        df = df[['token', 'pos', 'count']]
        df['verify'] = ''
        all_tokens = [item['token'] for i, item in df.iterrows()]

        tokens = [token for token in all_tokens if not vlsp_cache.contains(token)]
        MAX_ITEMS = 500
        words_data = VLSPDictionary.lookups(tokens[:MAX_ITEMS], n_workers=30)
        for token, word in zip(tokens, words_data):
            vlsp_cache.add(token, word)

        words = [vlsp_cache.get(token) for token in all_tokens]

        def check_has_tag(data):
            word, items = data
            tag = items[1]['pos']
            if word is None:
                return '', ''
            for sense in word.senses:
                if sense.tag == tag:
                    return 'x', sense.description
            return '', ''

        vlsp_data = [check_has_tag(data) for data in zip(words, df.iterrows())]
        votes, descriptions = zip(*vlsp_data)
        df['vlsp_votes'] = votes
        df['vlsp_description'] = descriptions
        file = join(DICTIONARY_FOLDER, "data", f"words_{pos_full_name}_candidates.xlsx")
        df.to_excel(file, index=False)
    vlsp_cache.save(vlsp_cache_file)



# flake8: noqa

from functools import partial
from typing import Deque

import haiku as hk
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import optax
from tqdm.auto import tqdm
from vietTTS.nat.config import DurationInput

from .config import FLAGS
from .data_loader import textgrid_data_loader
from .model import DurationModel
from .utils import load_latest_ckpt, print_flags, save_ckpt


def loss_fn(params, aux, rng, x: DurationInput, is_training=True):
    @hk.transform_with_state
    def net(x):
        return DurationModel(is_training=is_training)(x)

    durations, aux = net.apply(params, aux, rng, x)
    mask = jnp.arange(0, x.phonemes.shape[1])[None, :] < x.lengths[:, None]
    # NOT predict [WORD END] token
    mask = jnp.where(x.phonemes == FLAGS.word_end_index, False, mask)
    masked_loss = jnp.abs(durations - x.durations) * mask
    loss = jnp.sum(masked_loss) / jnp.sum(mask)
    return loss, aux


forward_fn = jax.jit(
    hk.transform_with_state(lambda x: DurationModel(is_training=False)(x)).apply
)


def predict_duration(params, aux, rng, x: DurationInput):
    d, _ = forward_fn(params, aux, rng, x)
    return d, x.durations


val_loss_fn = jax.jit(partial(loss_fn, is_training=False))

loss_vag = jax.value_and_grad(loss_fn, has_aux=True)

optimizer = optax.chain(
    optax.clip_by_global_norm(FLAGS.max_grad_norm),
    optax.adamw(FLAGS.duration_learning_rate, weight_decay=FLAGS.weight_decay),
)


@jax.jit
def update(params, aux, rng, optim_state, inputs: DurationInput):
    rng, new_rng = jax.random.split(rng)
    (loss, new_aux), grads = loss_vag(params, aux, rng, inputs)
    updates, new_optim_state = optimizer.update(grads, optim_state, params)
    new_params = optax.apply_updates(params, updates)
    return loss, (new_params, new_aux, new_rng, new_optim_state)


def initial_state(batch):
    rng = jax.random.PRNGKey(42)
    params, aux = hk.transform_with_state(lambda x: DurationModel(True)(x)).init(
        rng, batch
    )
    optim_state = optimizer.init(params)
    return params, aux, rng, optim_state


def plot_val_duration(step: int, batch, params, aux, rng):
    fn = FLAGS.ckpt_dir / f"duration_{step}.png"
    predicted_dur, gt_dur = predict_duration(params, aux, rng, batch)
    L = batch.lengths[0]
    x = np.arange(0, L) * 3
    plt.plot(predicted_dur[0, :L])
    plt.plot(gt_dur[0, :L])
    plt.legend(["predicted", "gt"])
    plt.title("Phoneme durations")
    plt.savefig(fn)
    plt.close()


def train():
    train_data_iter = textgrid_data_loader(
        FLAGS.data_dir, FLAGS.max_phoneme_seq_len, FLAGS.batch_size, mode="train"
    )
    val_data_iter = textgrid_data_loader(
        FLAGS.data_dir, FLAGS.max_phoneme_seq_len, FLAGS.batch_size, mode="val"
    )
    losses = Deque(maxlen=1000)
    val_losses = Deque(maxlen=100)
    latest_ckpt = load_latest_ckpt(FLAGS.ckpt_dir)
    if latest_ckpt is not None:
        last_step, params, aux, rng, optim_state = latest_ckpt
    else:
        last_step = -1
        print("Generate random initial states...")
        params, aux, rng, optim_state = initial_state(next(train_data_iter))

    tr = tqdm(
        range(last_step + 1, 1 + FLAGS.num_training_steps),
        total=1 + FLAGS.num_training_steps,
        initial=last_step + 1,
        ncols=80,
        desc="training",
    )
    for step in tr:
        batch = next(train_data_iter)
        loss, (params, aux, rng, optim_state) = update(
            params, aux, rng, optim_state, batch
        )
        losses.append(loss)

        if step % 10 == 0:
            val_loss, _ = val_loss_fn(params, aux, rng, next(val_data_iter))
            val_losses.append(val_loss)

        if step % 1000 == 0:
            loss = sum(losses).item() / len(losses)
            val_loss = sum(val_losses).item() / len(val_losses)
            plot_val_duration(step, next(val_data_iter), params, aux, rng)
            tr.write(
                f" {step:>6d}/{FLAGS.num_training_steps:>6d} | train loss {loss:.5f} | val loss {val_loss:.5f}"
            )
            save_ckpt(step, params, aux, rng, optim_state, ckpt_dir=FLAGS.ckpt_dir)


if __name__ == "__main__":
    print_flags(FLAGS.__dict__)
    if not FLAGS.ckpt_dir.exists():
        print("Create checkpoint dir at", FLAGS.ckpt_dir)
        FLAGS.ckpt_dir.mkdir(parents=True, exist_ok=True)
    train()



PRETRAINED = {
    'vi-dp-v1': 'https://github.com/undertheseanlp/underthesea/releases/download/resources/vi-dp-v1.zip',
    'vi-dp-v1a0': 'https://github.com/undertheseanlp/underthesea/releases/download/resources/vi-dp-v1a0.zip',
    'vi-dp-v1a1': 'https://github.com/undertheseanlp/underthesea/releases/download/resources/vi-dp-v1a1.zip',
    'biaffine-dep-en': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ptb.biaffine.dependency.char.zip',
    'biaffine-dep-zh': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ctb7.biaffine.dependency.char.zip',
    'biaffine-dep-bert-en': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ptb.biaffine.dependency.bert.zip',
    'biaffine-dep-bert-zh': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ctb7.biaffine.dependency.bert.zip',
    'crfnp-dep-en': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ptb.crfnp.dependency.char.zip',
    'crfnp-dep-zh': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ctb7.crfnp.dependency.char.zip',
    'crf-dep-en': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ptb.crf.dependency.char.zip',
    'crf-dep-zh': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ctb7.crf.dependency.char.zip',
    'crf2o-dep-en': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ptb.crf2o.dependency.char.zip',
    'crf2o-dep-zh': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ctb7.crf2o.dependency.char.zip',
    'crf-con-en': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ptb.crf.constituency.char.zip',
    'crf-con-zh': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ctb7.crf.constituency.char.zip',
    'crf-con-bert-en': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ptb.crf.constituency.bert.zip',
    'crf-con-bert-zh': 'https://github.com/yzhangcs/parser/releases/download/v1.0.0/ctb7.crf.constituency.bert.zip'
}



# -*- coding: utf-8 -*-
import pickle
import string
from os.path import join, dirname

from nltk import PunktSentenceTokenizer

sentence_tokenizer = None


def _load_model():
    global sentence_tokenizer
    if sentence_tokenizer is not None:
        return
    model_path = join(dirname(__file__), 'st_kiss-strunk-2006_2019_01_13.pkl')
    with open(model_path, 'rb') as fs:
        punkt_param = pickle.load(fs)

    punkt_param.sent_starters = {}
    abbrev_types = ['g.m.t', 'e.g', 'dr', 'dr', 'vs', "000", 'mr', 'mrs', 'prof', 'inc', 'tp', 'ts', 'ths',
                    'th', 'vs', 'tp', 'k.l', 'a.w.a.k.e', 't', 'a.i', '</i', 'g.w',
                    'ass',
                    'u.n.c.l.e', 't.e.s.t', 'ths', 'd.c', 've…', 'ts', 'f.t', 'b.b', 'z.e', 's.g', 'm.p',
                    'g.u.y',
                    'l.c', 'g.i', 'j.f', 'r.r', 'v.i', 'm.h', 'a.s', 'bs', 'c.k', 'aug', 't.d.q', 'b…', 'ph',
                    'j.k', 'e.l', 'o.t', 's.a']
    abbrev_types.extend(string.ascii_uppercase)
    for abbrev_type in abbrev_types:
        punkt_param.abbrev_types.add(abbrev_type)
    for abbrev_type in string.ascii_lowercase:
        punkt_param.abbrev_types.add(abbrev_type)
    sentence_tokenizer = PunktSentenceTokenizer(punkt_param)


def sent_tokenize(text):
    _load_model()
    sentences = sentence_tokenizer.sentences_from_text(text)
    return sentences



# -*- coding: utf-8 -*-
from underthesea import pos_tag
import sys

if sys.version_info >= (3, 0):
    from .model_crf import CRFChunkingPredictor
else:
    from model_crf import CRFChunkingPredictor


def chunk(sentence, format=None):
    """
    Vietnamese chunking

    Parameters
    ==========

    sentence: {unicode, str}
        raw sentence

    Returns
    =======
    tokens: 	list of tuple with word, pos tag, chunking tag
        tagged sentence

    Examples
    --------

    >>> # -*- coding: utf-8 -*-
    >>> from underthesea import chunk
    >>> sentence = "Nghi vấn 4 thi thể Triều Tiên trôi dạt bờ biển Nhật Bản"
    >>> chunk(sentence)
    [('Nghi vấn', 'N', 'B-NP'),
    ('4', 'M', 'B-NP'),
    ('thi thể', 'N', 'B-NP'),
    ('Triều Tiên', 'Np', 'B-NP'),
    ('trôi dạt', 'V', 'B-VP'),
    ('bờ biển', 'N', 'B-NP'),
    ('Nhật Bản', 'Np', 'B-NP')]
    """
    sentence = pos_tag(sentence)
    crf_model = CRFChunkingPredictor.Instance()
    result = crf_model.predict(sentence, format)
    return result



# flake8: noqa

import pickle
from argparse import ArgumentParser
from pathlib import Path

import haiku as hk
import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np

from .config import FLAGS, DurationInput
from .data_loader import load_phonemes_set
from .model import AcousticModel, DurationModel


def load_lexicon(fn):
    lines = open(fn, "r").readlines()
    lines = [l.lower().strip().split("\t") for l in lines]
    return dict(lines)


def predict_duration(tokens, ckpt_file):
    def fwd_(x):
        return DurationModel(is_training=False)(x)

    forward_fn = jax.jit(hk.transform_with_state(fwd_).apply)
    with open(ckpt_file, "rb") as f:
        dic = pickle.load(f)
    x = DurationInput(
        np.array(tokens, dtype=np.int32)[None, :],
        np.array([len(tokens)], dtype=np.int32),
        None,
    )
    return forward_fn(dic["params"], dic["aux"], dic["rng"], x)[0]


def text2tokens(text, lexicon_fn):
    phonemes = load_phonemes_set()
    lexicon = load_lexicon(lexicon_fn)

    words = text.strip().lower().split()
    tokens = [FLAGS.sil_index]
    for word in words:
        if word in FLAGS.special_phonemes:
            tokens.append(phonemes.index(word))
        elif word in lexicon:
            p = lexicon[word]
            p = p.split()
            p = [phonemes.index(pp) for pp in p]
            tokens.extend(p)
            tokens.append(FLAGS.word_end_index)
        else:
            for p in word:
                if p in phonemes:
                    tokens.append(phonemes.index(p))
            tokens.append(FLAGS.word_end_index)
    tokens.append(FLAGS.sil_index)  # silence
    return tokens

dic = None
def predict_mel(tokens, durations, ckpt_fn):
    global dic
    with open(ckpt_fn, "rb") as f:
        if dic is None:
            dic = pickle.load(f)
        last_step, params, aux, rng, optim_state = (
            dic["step"],
            dic["params"],
            dic["aux"],
            dic["rng"],
            dic["optim_state"],
        )

    @hk.transform_with_state
    def forward(tokens, durations, n_frames):
        net = AcousticModel(is_training=False)
        return net.inference(tokens, durations, n_frames)

    durations = durations * FLAGS.sample_rate / (FLAGS.n_fft // 4)
    n_frames = int(jnp.sum(durations).item())
    predict_fn = jax.jit(forward.apply, static_argnums=[5])
    tokens = np.array(tokens, dtype=np.int32)[None, :]
    return predict_fn(params, aux, rng, tokens, durations, n_frames)[0]


def text2mel(
    text: str,
    lexicon_fn=FLAGS.data_dir / "lexicon.txt",
    silence_duration: float = -1.0,
    acoustic_ckpt=FLAGS.ckpt_dir / "acoustic_latest_ckpt.pickle",
    duration_ckpt=FLAGS.ckpt_dir / "duration_latest_ckpt.pickle",
):
    tokens = text2tokens(text, lexicon_fn)
    durations = predict_duration(tokens, duration_ckpt)
    durations = jnp.where(
        np.array(tokens)[None, :] == FLAGS.sil_index,
        jnp.clip(durations, a_min=silence_duration, a_max=None),
        durations,
    )
    durations = jnp.where(
        np.array(tokens)[None, :] == FLAGS.word_end_index, 0.0, durations
    )
    mels = predict_mel(tokens, durations, acoustic_ckpt)
    if tokens[-1] == FLAGS.sil_index:
        end_silence = durations[0, -1].item()
        silence_frame = int(end_silence * FLAGS.sample_rate / (FLAGS.n_fft // 4))
        mels = mels[:, : (mels.shape[1] - silence_frame)]
    return mels


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--text", type=str, required=True)
    parser.add_argument("--output", type=Path, required=True)
    args = parser.parse_args()
    mel = text2mel(args.text)
    plt.figure(figsize=(10, 5))
    plt.imshow(mel[0].T, origin="lower", aspect="auto")
    plt.savefig(str(args.output))
    plt.close()
    mel = jax.device_get(mel)
    mel.tofile("clip.mel")



import sys
from os import listdir
from os.path import join, basename
import charset_normalizer

from underthesea.feature_engineering.text import Text
from underthesea.file_utils import DATASETS_FOLDER
from underthesea.pipeline.word_tokenize import tokenize

SUPPORTED_CORPUS_TYPE = set(["TOKENIZE"])

error_count = 0
DEFAULT_MAX_ERROR = 30
MAX_ERROR = DEFAULT_MAX_ERROR


def warn(message, level=1, file=None, line_number=None, error_type=None, sent_id=None, node_number=None):
    global error_count
    error_count += 1
    if error_count >= MAX_ERROR:
        print("MAX_ERROR_EXCEEDED. Stop")
        print(f"*** FAILED *** with {error_count} errors")
        sys.exit(1)

    text = ""
    if file:
        text += f"[(in {file})"
    if line_number:
        text += f" Line {line_number}"
    if sent_id:
        text += f" Sent {sent_id}"
    text += "]: "
    text += f"[L{level}"
    if error_type:
        text += f" {error_type}"
    text += "] "
    text += message
    print(text)


def validate_corpus_exist(corpus_name):
    if corpus_name not in listdir(DATASETS_FOLDER):
        print(f"Corpus {corpus_name} is not in existed in {DATASETS_FOLDER}")
        sys.exit(1)


def validate_corpus_type(corpus_type):
    if corpus_type not in SUPPORTED_CORPUS_TYPE:
        print(f"{corpus_type} is not supported")
        sys.exit(1)


# ======================================================================================================================
# LEVEL 1
# TODO: Check NFC Normalize
# ======================================================================================================================
def validate_utf8(file):
    base_name = basename(file)
    text = b''
    with open(file, "rb") as f:
        for i, line in enumerate(f):
            if i < 1000:
                text += line
            else:
                break
    results = charset_normalizer.from_bytes(text).best()
    if not (results.encoding == "utf-8" and results.coherence >= 0.99):
        warn(message=f"File {file} should encoding with UTF-8", level=1)
        sys.exit(1)
    with open(file, "r") as f:
        content = f.read()
    normalized_nfc_content = Text(content)
    if normalized_nfc_content != content:
        warn(message=f"File {base_name} should normalized to NFC",
             error_type="Format nfc-normalized-failed",
             file=base_name, level=1)


# ======================================================================================================================
# LEVEL 2
# ======================================================================================================================
def fetch_sentence(file):
    sentence = ""
    with open(file) as f:
        for i, line in enumerate(f):
            if line == "\n":
                yield i, sentence
                sentence = ""
            else:
                sentence += line


def validate_sentence_format(sentence, i_start, file):
    base_name = basename(file)
    nodes = sentence.strip().split("\n")
    comment_nodes = [node for node in nodes if node.startswith("#")]
    has_sent_id = False
    has_text = False
    for comment_node in comment_nodes:
        if comment_node.startswith("# sent_id = "):
            has_sent_id = True
        if comment_node.startswith("# text = "):
            has_text = True
    if not has_sent_id:
        warn(message="Sentence must has sent_id",
             error_type="Format no-sent_id",
             file=base_name, line_number=i_start, level=2)
    if not has_text:
        warn(message="Sentence must has text",
             error_type="Format no-text",
             file=base_name, line_number=i_start, level=2)


# ======================================================================================================================
# LEVEL 3
# ======================================================================================================================
def validate_token(sentence, i_end, file):
    base_name = basename(file)
    nodes = sentence.strip().split("\n")
    i_start = i_end - len(nodes) + 1
    sent_id = None
    for i, node in enumerate(nodes):
        if node.startswith("#"):
            if node.startswith("# sent_id = "):
                sent_id = node[12:]
            continue
        tokens = node.split("\t")
        line_number = i_start + i
        if len(tokens) != 2:
            warn(message="Each row must has 2 tokens",
                 error_type="Format invalid-num-tokens",
                 file=base_name, line_number=line_number, level=2)
        token = tokens[0]
        tag = tokens[1]
        if token.startswith(" ") or token.endswith(" "):
            warn(message=f"token should not start and end with spaces: {token}",
                 error_type="Format token-error",
                 file=base_name, line_number=line_number, level=3)

        if tag not in ["B-W", "I-W"]:
            warn(message=f"Tag must be B-W or I-W, found {tag}",
                 error_type="Format tag-error",
                 file=base_name, line_number=line_number, level=3, sent_id=sent_id)

    content_nodes = [node for node in nodes if not node.startswith("#")]
    content_tokens = [node.split("\t")[0] for node in content_nodes]
    text = " ".join(content_tokens)
    tokenized_tokens = tokenize(text)
    if tokenized_tokens != content_tokens:
        if len(text) > 53:
            message = "tokenized error for text: " + text[:50] + "..."
        else:
            message = "tokenized error for text: " + text
        message += '\n'
        message += '  Actual:' + str(content_tokens)
        message += '\n'
        message += 'Expected:' + str(tokenized_tokens)
        warn(message=message,
             error_type="Format tokenize-error",
             file=base_name, line_number=i_start, level=3, sent_id=sent_id)


def validate_content(file):
    for i, sentence in fetch_sentence(file):
        validate_sentence_format(sentence, i, file)
        validate_token(sentence, i, file)


def validate_corpus_content(corpus_name):
    # Level 1
    # Validate NFC Encoding
    corpus_folder = join(DATASETS_FOLDER, corpus_name)
    train = join(corpus_folder, "train.txt")
    test = join(corpus_folder, "test.txt")
    files = [train, test]
    for file in files:
        validate_utf8(file)

    # Level 2, 3
    # Check content
    for file in files:
        validate_content(file)


def validate_corpus(corpus_type, corpus_name, max_error=DEFAULT_MAX_ERROR):
    global error_count
    global MAX_ERROR
    MAX_ERROR = max_error
    error_count = 0
    print(f"Validate {corpus_type} corpus: {corpus_name}")
    validate_corpus_type(corpus_type)
    validate_corpus_content(corpus_name)

    if error_count == 0:
        print("*** PASSED ***")
    else:
        print(f"*** FAILED *** with {error_count} errors")



from pydub import AudioSegment

if __name__ == '__main__':
    sound1 = AudioSegment.from_wav("sound.wav")
    sound2 = AudioSegment.from_wav("chill.wav")

    # overlay sound2 over sound1 at the 5th second mark
    combined_sounds = sound1.overlay(sound2)

    combined_sounds.export("combined_sounds.wav", format="wav")
